"""
Reasoner engine.

This module provides the implementation of the
ReasonerEngine class, which is used to perform
reasoning tasks using a Language Model (LLM).
The engine takes an ontology and a query task as input,
translates it into an LLM prompt, and performs
reasoning over the ontology.

Classes:
    ReasonerResult: Represents the result of a reason query.
    ReasonerResultSet: Represents a set of reasoner results.
    ReasonerEngine: Engine for performing reasoning using an LLM.

Functions:
    flatten_list(lst): Flattens a nested list into a single list.
    reason(task, template_path=None, strict=False, evaluate=None): Performs
    reasoning over axioms and query entailments.
    reason_multiple(task_collection, **kwargs): Performs reasoning over multiple tasks.
    _parse_single_answer(payload, task): Parses a single answer from the payload.
    evaluate(result, task): Evaluates the result against the task.

Usage:
    The ReasonerEngine can be used to perform reasoning tasks
    such as finding superclasses of a given class.
    It can also provide explanations for the reasoning process.
"""

import logging
import re
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union

from jinja2 import Template
from pydantic import BaseModel

from ontogpt.engines.knowledge_engine import KnowledgeEngine
from ontogpt.ontex.extractor import (
    Answer,
    Axiom,
    Explanation,
    LLMReasonMethodType,
    Task,
    TaskCollection,
)
from ontogpt.prompts.reasoning import DEFAULT_REASONING_PROMPT
from ontogpt.utils.parse_utils import split_on_one_of

logger = logging.getLogger(__name__)


def flatten_list(lst):
    flat_list = []
    for item in lst:
        if isinstance(item, list):
            flat_list.extend(flatten_list(item))
        else:
            flat_list.append(item)
    return flat_list


class ReasonerResult(BaseModel):
    """The result of a reason query."""

    name: Optional[str] = None
    completed: Optional[bool] = True
    task_name: Optional[str] = None
    task_type: Optional[str] = None
    task_obfuscated: Optional[bool] = None
    method: Optional[LLMReasonMethodType] = None
    model: Optional[str] = None
    description: Optional[str] = None
    answers: Optional[List[Answer]] = None
    prompt: Optional[str] = None
    completion: Optional[str] = None
    jaccard_score: Optional[float] = 0.0
    false_positives: Optional[List[str]] = None
    false_negatives: Optional[List[str]] = None
    num_false_positives: Optional[int] = None
    num_false_negatives: Optional[int] = None
    num_true_positives: Optional[int] = None
    num_true_negatives: Optional[int] = None
    precision: Optional[float] = 0.0
    recall: Optional[float] = None
    f1_score: Optional[float] = None
    len_shortest_explanation: Optional[int] = None

    class Config:
        """Pydantic config."""

        use_enum_values = True


class ReasonerResultSet(BaseModel):
    name: str = ""
    results: List[ReasonerResult]


@dataclass
class ReasonerEngine(KnowledgeEngine):
    """Engine for performing reasoning using an LLM.

    This engine takes as input an Ontology, and a query Task,
    and then translates this to an LLM prompt that asks the LLM to
    perform the task over the ontology after reasoning over it.

    The Task is typically a query such as finding superclasses of
    a given class.

    This is intended primarily for investigation purposes. For practical
    scenarios, it is recommended to use a dedicated OWL reasoner. The goal
    of this engine is to evaluate the extent to which the LLM can perform
    reasoning-like tasks, including deduction and abduction (explanation).

    Due to token-length constraints on some models, it is usually necessary
    to extract a submodule prior to reasoning. This can be done using the
    OntologyExtractor:

    >>> from oaklib import get_adapter
    >>> from ontogpt.ontex.extractor import OntologyExtractor, Task
    >>> adapter = get_adapter("sqlite:obo:go")
    >>> extractor = OntologyExtractor(adapter=adapter)
    >>> task = extractor.extract_indirect_superclasses_task(
    ...    subclass="GO:0005634", siblings=["GO:0005773"], roots=["GO:0043226"]
    ... )

    The extractor will actually perform the intended reasoning task, and include
    the expected answer in the Task object. This may seem a little circular, but
    the goal here is to evaluate.

    The task can then be passed to the reasoner engine:

    >>> from ontogpt.engines.reasoner_engine import ReasonerEngine
    >>> reasoner = ReasonerEngine()
    >>> result = reasoner.reason(task)

    We can expand on this:

    >>> for answer in result.answers:
    ...    print(f"PARENT: {answer.text}")
    <BLANKLINE>
    PARENT: MembraneBoundedOrganelle
    ...
    PARENT: IntracellularOrganelle

    We can also ask for explanations:

    >>> task.include_explanations = True
    >>> result = reasoner.reason(task)
    >>> for answer in result.answers:
    ...    print(f"PARENT: {answer.text}")
    ...    if answer.explanations:
    ...        print("  EXPLANATION:")
    ...        for e in answer.explanations:
    ...            for axiom in e.axioms:
    ...                print(f"    AXIOM: {axiom.text}")
    <BLANKLINE>
    ...
    PARENT: IntracellularOrganelle
        EXPLANATION:
            AXIOM: Nucleus SubClassOf IntracellularMembraneBoundedOrganelle
            AXIOM: IntracellularMembraneBoundedOrganelle SubClassOf IntracellularOrganelle

    """

    completion_length = 250

    def reason(
        self, task: Task, template_path=None, strict=False, evaluate: bool = None
    ) -> ReasonerResult:
        """Reason over axioms and query entailments."""
        if template_path is None:
            template_path = DEFAULT_REASONING_PROMPT
        if isinstance(template_path, Path):
            template_path = str(template_path)
        if isinstance(template_path, str):
            # create a Jinja2 template object
            with open(template_path) as file:
                template_txt = file.read()
                template = Template(template_txt)
        prompt = template.render(
            task=task,
            ontology=task.ontology,
            query=task.query,
            examples=task.examples,
        )
        completion_length = self.completion_length
        if task.method == LLMReasonMethodType.EXPLANATION:
            completion_length *= 2
        elif task.method == LLMReasonMethodType.CHAIN_OF_THOUGHT:
            completion_length *= 2
        logger.info(f"Prompt: {prompt}")
        if self.encoding is not None:
            prompt_length = len(self.encoding.encode(prompt)) + 10
        else:
            prompt_length = len(prompt)
        completed = True
        logger.info(f"PROMPT LENGTH: {prompt_length}")

        payload = self.client.complete(prompt, max_tokens=completion_length)
        if task.has_multiple_answers:
            elements = payload.split("- ")
            answers = [self._parse_single_answer(e, task) for e in elements]
        else:
            answers = [self._parse_single_answer(payload, task)]
        answers = [a for a in flatten_list(answers) if a is not None]
        result = ReasonerResult(
            completed=completed,
            task_name=task.name,
            task_type=task.type,
            task_obfuscated=task.obfuscated,
            method=task.method,
            len_shortest_explanation=task.len_shortest_explanation,
            model=self.model,
            prompt=prompt,
            completion=payload,
        )
        # TODO: determine which it doesn't work to initialize with this
        result.answers = answers  # type: ignore
        logger.debug(f"Answers: {task.answers} // {answers}")
        result.name = f"{task.name}-{task.method.value}-{self.model}"
        if not task.answers and evaluate:
            raise ValueError(f"Cannot evaluate without expected answers: {task}")
        if task.answers is not None:
            self.evaluate(result, task)
        return result

    def reason_multiple(self, task_collection: TaskCollection, **kwargs) -> ReasonerResultSet:
        """
        Reason over multiple tasks.

        :param task_collection:
        :param kwargs:
        :return:
        """
        results = [self.reason(task, **kwargs) for task in task_collection.tasks]
        return ReasonerResultSet(results=results)

    def _parse_single_answer(
        self, payload: str, task: Task
    ) -> Optional[Union[Answer, List[Answer]]]:
        """Parse single answer from payload.

        In some cases for COT reasoning the model may compress multiple
        answers into a single answer line.
        """
        payload = payload.strip()
        if not payload:
            return None
        payload = payload.replace("\n", " ")
        if task.chain_of_thought:
            pattern = r"^REASONING:\s*\[\s*(.*?)\s*\]\s*CONCLUSION:\s*(.*?)$"
            match = re.match(pattern, payload)
            print(f"CHAIN OF THOUGHT: {payload}")
            if match:
                print(f"MATCH: {match.groups()}")
                explanation = match.group(1)
                text = match.group(2)
                # rest = match.group(3)
                axioms = [x.strip() for x in split_on_one_of(explanation, [";", ","])]
                explanation = Explanation(axioms=[Axiom(text=a) for a in axioms if a])
                axiom_texts = split_on_one_of(text, [";", ","])
                if len(axiom_texts) == 1:
                    return Answer(text=text, explanations=[explanation])
                else:
                    return [
                        Answer(text=axiom_text, explanations=[explanation])
                        for axiom_text in axiom_texts
                    ]
        pattern = r"^(.*?)\s*\[\s*(.*?)\s*\]\s*(.*?)$"
        match = re.match(pattern, payload)
        if match:
            text = match.group(1)
            explanation = match.group(2)
            rest = match.group(3)
            axioms = [x.strip() for x in split_on_one_of(explanation, [";", ","])]
            explanation = Explanation(axioms=[Axiom(text=a) for a in axioms if a])
            if rest:
                explanation.comments = [rest]
            return Answer(text=text, explanations=[explanation])
        else:
            return Answer(text=payload)

    def evaluate(self, result: ReasonerResult, task: Task):
        """Evaluate result against task."""
        logger.debug(f"Evaluating result: {result}")
        positives = {t.text for t in task.answers}
        result_answer_texts = {a.text for a in result.answers}
        ixn = positives.intersection(result_answer_texts)
        all_texts = positives.union(result_answer_texts)
        result.false_positives = list(result_answer_texts - positives)
        result.false_negatives = list(positives - result_answer_texts)
        result.num_false_positives = len(result.false_positives)
        result.num_false_negatives = len(result.false_negatives)
        result.num_true_positives = len(ixn)
        tp_plus_tn = result.num_true_positives + result.num_false_positives
        if tp_plus_tn == 0:
            result.precision = 0.0
        else:
            result.precision = result.num_true_positives / tp_plus_tn
        result.recall = result.num_true_positives / len(positives)
        if len(all_texts) == 0:
            result.jaccard_score = 0.0
        else:
            result.jaccard_score = len(ixn) / len(all_texts)
        if result.num_true_positives == 0:
            result.f1_score = 0.0
        else:
            result.f1_score = (
                2 * (result.precision * result.recall) / (result.precision + result.recall)
            )



"""
Embedding similarity engine.

This module provides functionality for calculating the similarity between entities
using embeddings and cosine similarity. It includes classes and methods for
searching for similar entities and retrieving textual representations of entities.

Classes:
    EmbeddingSimilarity: A dataclass representing the similarity between two entities.
    SimilarityEngine: A class for calculating similarity between entities.

Methods:
    SimilarityEngine.similarity(entity1: str, entity2: str) -> EmbeddingSimilarity:
        Get similarity between two entities.
    SimilarityEngine.search(entity1: str, entities: List[str]) -> Iterable[EmbeddingSimilarity]:
        Search for similar entities.
    SimilarityEngine.entity_text(entity: str) -> str:
        Get text representation for an entity.
"""

import logging
from dataclasses import dataclass
from typing import Iterable, List

from oaklib import BasicOntologyInterface
from oaklib.datamodels.vocabulary import IS_A

from ontogpt.clients import LLMClient
from ontogpt.engines.knowledge_engine import KnowledgeEngine

logger = logging.getLogger(__name__)


@dataclass
class EmbeddingSimilarity:
    subject_id: str = ""
    subject_label: str = ""
    object_id: str = ""
    object_label: str = ""
    embedding_cosine_similarity: float = None
    object_rank_for_subject: int = None


@dataclass
class SimilarityEngine(KnowledgeEngine):
    """Engine for finding embedding similarity."""

    adapter: BasicOntologyInterface = None
    autolabel: bool = True
    definitions: bool = True
    parents: bool = True
    ancestors: bool = True
    synonyms: bool = True
    logical_definitions: bool = False

    def similarity(self, entity1: str, entity2: str) -> EmbeddingSimilarity:
        """Get similarity."""
        t1 = self.entity_text(entity1)
        t2 = self.entity_text(entity2)
        client = LLMClient()
        score = client.similarity(t1, t2)
        obj = EmbeddingSimilarity(
            subject_id=entity1, object_id=entity2, embedding_cosine_similarity=score
        )
        if self.autolabel:
            obj.subject_label = self.adapter.label(entity1)
            obj.object_label = self.adapter.label(entity2)
        return obj

    def search(self, entity1: str, entities: List[str]) -> Iterable[EmbeddingSimilarity]:
        """Get similarity."""
        client = LLMClient()
        t1 = self.entity_text(entity1)
        sims = []
        for entity2 in entities:
            t2 = self.entity_text(entity2)
            score = client.similarity(t1, t2)
            sim = EmbeddingSimilarity(
                subject_id=entity1, object_id=entity2, embedding_cosine_similarity=score
            )
            if self.autolabel:
                sim.subject_label = self.adapter.label(entity1)
                sim.object_label = self.adapter.label(entity2)
            sims.append(sim)
        sims = sorted(sims, key=lambda x: x.embedding_cosine_similarity, reverse=True)
        for i, sim in enumerate(sims):
            sim.object_rank_for_subject = i
        yield from sims

    def entity_text(self, entity: str) -> str:
        """Get text for an entity."""
        adapter = self.adapter
        s = f"{adapter.label(entity)}"
        if self.definitions:
            s += f"\ndefinition: {adapter.definition(entity)}"
        if self.parents:
            parent_labels = [
                adapter.label(o) for _s, _p, o in adapter.relationships([entity], [IS_A])
            ]
            s += f"\nparents: {'; '.join(parent_labels)}"
        if self.ancestors:
            ancestor_labels = [
                adapter.label(a) for a in adapter.ancestors([entity], [IS_A], reflexive=False)
            ]
            s += f"\nancestors: {'; '.join(ancestor_labels)}"
        if self.synonyms:
            s += f"\nsynonyms: {'; '.join(adapter.entity_aliases(entity))}"
        if self.logical_definitions:
            for ldef in adapter.logical_definitions(entity):
                genus_labels = [adapter.label(g) for g in ldef.genusIds]
                restriction_labels = [
                    f"{adapter.label(r.propertyId)} {adapter.label(r.valueId)}"
                    for r in ldef.restrictions
                ]
                s += f"\nlogical definition: A {', '.join(genus_labels)} that\
                    {' and '.join(restriction_labels)}"
            s += f"\nlogical definitions: {'; '.join(adapter.logical_definitions(entity))}"
        logger.info(f"Entity text for {entity}: {s}")
        return s



"""Utilities for working with clinical case reports.

The prompt in get_kanjee_prompt() is that used by
the following citation:
    Kanjee Z, Crowe B, Rodman A. Accuracy of a
    generative artificial intelligence model in
    a complex diagnostic challenge. JAMA.
    2023 Jul 3;330(1):78–80.
    http://dx.doi.org/10.1001/jama.2023.8288
    PMCID: PMC10273128
"""


def get_kanjee_prompt() -> str:
    """Prompt from Kanjee et al. 2023."""
    prompt = (
        "I am running an experiment on a clinicopathological case conference to see "
        "how your diagnoses compare with those of human experts. I am going to give "
        "you part of a medical case. These have all been published in the New England "
        "Journal of Medicine. You are not trying to treat any patients. As you read the "
        "case, you will notice that there are expert discussants giving their thoughts. "
        'In this case, you are "Dr. GPT-4," an Al language model who is discussing '
        "the case along with human experts. A clinicopathological case conference has "
        "several unspoken rules. The first is that there is most often a single definitive "
        "diagnosis (though rarely there may be more than one), and it is a diagnosis that "
        "is known today to exist in humans. The diagnosis is almost always confirmed by "
        "some sort of clinical pathology test or anatomic pathology test, though in "
        "rare cases when such a test does not exist for a diagnosis the diagnosis can "
        "instead be made using validated clinical criteria or very rarely just confirmed "
        "by expert opinion. You will be told at the end of the case description whether "
        "a diagnostic test/tests are being ordered, which you can assume will make the "
        "diagnosis/diagnoses. After you read the case, I want you to give two pieces of "
        "information. The first piece of information is your most likely diagnosis/diagnoses. "
        "You need to be as specific as possible -- the goal is to get the correct answer, "
        "not a broad category of answers. You do not need to explain your reasoning, just "
        "give the diagnosis/diagnoses. The second piece of information is to give a robust "
        "differential diagnosis, ranked by their probability so that the most likely "
        "diagnosis is at the top, and the least likely is at the bottom. There is no limit "
        "to the number of diagnoses on your differential. You can give as many diagnoses "
        "as you think are reasonable. You do not need to explain your reasoning, just list"
        " the diagnoses. Again, the goal is to be as specific as possible with each of the "
        "diagnoses. Do you have any questions, Dr. GPT-4?\n\nHere is the case:"
    )
    return prompt



from __future__ import annotations 
from datetime import (
    datetime,
    date
)
from decimal import Decimal 
from enum import Enum 
import re
import sys
from typing import (
    Any,
    ClassVar,
    List,
    Literal,
    Dict,
    Optional,
    Union
)
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
    field_validator
)
metamodel_version = "None"
version = "None"


class ConfiguredBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment = True,
        validate_default = True,
        extra = "forbid",
        arbitrary_types_allowed = True,
        use_enum_values = True,
        strict = False,
    )
    pass




class LinkMLMeta(RootModel):
    root: Dict[str, Any] = {}
    model_config = ConfigDict(frozen=True)

    def __getattr__(self, key:str):
        return getattr(self.root, key)

    def __getitem__(self, key:str):
        return self.root[key]

    def __setitem__(self, key:str, value):
        self.root[key] = value

    def __contains__(self, key:str) -> bool:
        return key in self.root


linkml_meta = LinkMLMeta({'default_prefix': 'food',
     'default_range': 'string',
     'description': 'A template for extracting food names and terms from text.',
     'id': 'http://w3id.org/ontogpt/food',
     'imports': ['linkml:types', 'core'],
     'license': 'https://creativecommons.org/publicdomain/zero/1.0/',
     'name': 'food',
     'prefixes': {'GO': {'prefix_prefix': 'GO',
                         'prefix_reference': 'http://purl.obolibrary.org/obo/GO_'},
                  'food': {'prefix_prefix': 'food',
                           'prefix_reference': 'http://w3id.org/ontogpt/food'},
                  'foodon': {'prefix_prefix': 'foodon',
                             'prefix_reference': 'http://purl.obolibrary.org/obo/foodon_'},
                  'linkml': {'prefix_prefix': 'linkml',
                             'prefix_reference': 'https://w3id.org/linkml/'},
                  'rdf': {'prefix_prefix': 'rdf',
                          'prefix_reference': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'}},
     'source_file': 'src/ontogpt/templates/food.yaml',
     'title': 'Food Extraction Template'} )

class NullDataOptions(str, Enum):
    UNSPECIFIED_METHOD_OF_ADMINISTRATION = "UNSPECIFIED_METHOD_OF_ADMINISTRATION"
    NOT_APPLICABLE = "NOT_APPLICABLE"
    NOT_MENTIONED = "NOT_MENTIONED"



class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    input_id: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'input_id', 'domain_of': ['ExtractionResult']} })
    input_title: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'input_title', 'domain_of': ['ExtractionResult']} })
    input_text: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'input_text', 'domain_of': ['ExtractionResult']} })
    raw_completion_output: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'raw_completion_output', 'domain_of': ['ExtractionResult']} })
    prompt: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'prompt', 'domain_of': ['ExtractionResult']} })
    extracted_object: Optional[Any] = Field(None, description="""The complex objects extracted from the text""", json_schema_extra = { "linkml_meta": {'alias': 'extracted_object', 'domain_of': ['ExtractionResult']} })
    named_entities: Optional[List[Any]] = Field(None, description="""Named entities extracted from the text""", json_schema_extra = { "linkml_meta": {'alias': 'named_entities', 'domain_of': ['ExtractionResult']} })


class NamedEntity(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'abstract': True, 'from_schema': 'http://w3id.org/ontogpt/core'})

    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'annotations': {'owl': {'tag': 'owl',
                                 'value': 'AnnotationProperty, AnnotationAssertion'}},
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })


class CompoundExpression(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'abstract': True, 'from_schema': 'http://w3id.org/ontogpt/core'})

    pass


class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'abstract': True, 'from_schema': 'http://w3id.org/ontogpt/core'})

    subject: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'subject', 'domain_of': ['Triple']} })
    predicate: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'predicate', 'domain_of': ['Triple']} })
    object: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'object', 'domain_of': ['Triple']} })
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""", json_schema_extra = { "linkml_meta": {'alias': 'qualifier', 'domain_of': ['Triple']} })
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""", json_schema_extra = { "linkml_meta": {'alias': 'subject_qualifier', 'domain_of': ['Triple']} })
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""", json_schema_extra = { "linkml_meta": {'alias': 'object_qualifier', 'domain_of': ['Triple']} })


class TextWithTriples(ConfiguredBaseModel):
    """
    A text containing one or more relations of the Triple type.
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    publication: Optional[Publication] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'publication',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'domain_of': ['TextWithTriples', 'TextWithEntity']} })
    triples: Optional[List[Triple]] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'triples', 'domain_of': ['TextWithTriples']} })


class TextWithEntity(ConfiguredBaseModel):
    """
    A text containing one or more instances of a single type of entity.
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    publication: Optional[Publication] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'publication',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'domain_of': ['TextWithTriples', 'TextWithEntity']} })
    entities: Optional[List[str]] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'entities', 'domain_of': ['TextWithEntity']} })


class RelationshipType(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core',
         'id_prefixes': ['RO', 'biolink']})

    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'annotations': {'owl': {'tag': 'owl',
                                 'value': 'AnnotationProperty, AnnotationAssertion'}},
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })


class Publication(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    id: Optional[str] = Field(None, description="""The publication identifier""", json_schema_extra = { "linkml_meta": {'alias': 'id', 'domain_of': ['NamedEntity', 'Publication']} })
    title: Optional[str] = Field(None, description="""The title of the publication""", json_schema_extra = { "linkml_meta": {'alias': 'title', 'domain_of': ['Publication']} })
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""", json_schema_extra = { "linkml_meta": {'alias': 'abstract', 'domain_of': ['Publication']} })
    combined_text: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'combined_text', 'domain_of': ['Publication']} })
    full_text: Optional[str] = Field(None, description="""The full text of the publication""", json_schema_extra = { "linkml_meta": {'alias': 'full_text', 'domain_of': ['Publication']} })


class AnnotatorResult(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    subject_text: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'subject_text', 'domain_of': ['AnnotatorResult']} })
    object_id: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'object_id', 'domain_of': ['AnnotatorResult']} })
    object_text: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'object_text', 'domain_of': ['AnnotatorResult']} })


class FoodSet(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/food', 'tree_root': True})

    terms: Optional[List[str]] = Field(None, description="""A semicolon-separated list of any names of foods.""", json_schema_extra = { "linkml_meta": {'alias': 'terms', 'domain_of': ['FoodSet']} })
    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'annotations': {'owl': {'tag': 'owl',
                                 'value': 'AnnotationProperty, AnnotationAssertion'}},
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })


class FoodTerm(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'annotations': {'annotators': {'tag': 'annotators',
                                        'value': 'sqlite:obo:foodon'},
                         'prompt': {'tag': 'prompt',
                                    'value': 'The name of a food. Examples include: '
                                             'apple juice, okra pod, chocolate '
                                             'substitute, breakfast cereal, tuna '
                                             '(flaked, canned), beef chuck roast'}},
         'from_schema': 'http://w3id.org/ontogpt/food',
         'id_prefixes': ['FOODON']})

    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'annotations': {'owl': {'tag': 'owl',
                                 'value': 'AnnotationProperty, AnnotationAssertion'}},
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
ExtractionResult.model_rebuild()
NamedEntity.model_rebuild()
CompoundExpression.model_rebuild()
Triple.model_rebuild()
TextWithTriples.model_rebuild()
TextWithEntity.model_rebuild()
RelationshipType.model_rebuild()
Publication.model_rebuild()
AnnotatorResult.model_rebuild()
FoodSet.model_rebuild()
FoodTerm.model_rebuild()




"""Parser utilities for handling output."""

import time
from oaklib import get_adapter

# TODO: Set this up to be more general
# TODO: merge the null checker into the knowledge engine

NULL_VALS = [
    "",
    "Not mentioned",
    "none mentioned",
    "Not mentioned in the text",
    "Not mentioned in the provided text.",
    "No exposures mentioned in the text.",
    "None",
    "N/A",
    "No exposures mentioned in the text.",
    "None mentioned in the text",
    "Not mentioned in the text.",
    "None relevant",
    "None mentioned in the text.",
    "No gene to molecular activity relationships mentioned in the text.",
    "No genes mentioned",
    "No genes mentioned in the text.",
]
lines = []
adapter = get_adapter("sqlite:obo:MONDO")
output_file = "output_partial_50_0911.yaml"


def tripleprint(dict):
    key = next(iter(dict))
    for i in range(len(dict[key])):
        curr = []
        for value in dict.values():
            curr.append(value[i])
        print(curr)


def forprint(lst):
    for elem in lst:
        print(elem)


def nprint(lst, n):
    for i in range(n):
        print(lst[i])


def sleepprint(lst):
    for elem in lst:
        print(elem)
        time.sleep(0.5)


def enumprint(lst):
    for index, elem in enumerate(lst):
        print(str(index + 1) + ":\t" + str(elem))


with open(output_file, "r") as file:
    to_print = False
    perpetuators = tuple(["  subject:", "  predicate:", "  object:", "    "])
    for line in file:
        line = line.strip("\n")
        if line.startswith("extracted_object:"):
            to_print = True
        elif not line.startswith(perpetuators):
            to_print = False
        if to_print:
            lines.append(line)

cleaned_lines = [x for x in list(filter(lambda elem: not (elem.isspace()), lines)) if x.strip()]
cleaned_lines = [x for x in cleaned_lines if x != "extracted_object: {}"]
i = 1
while i < len(cleaned_lines):
    if cleaned_lines[i].startswith("    "):
        cleaned_lines[i - 1] += " "
        cleaned_lines[i - 1] += cleaned_lines[i][4:]
        del cleaned_lines[i]
    else:
        i += 1

header = "extracted_object"
i = 0
while i < len(cleaned_lines):
    if cleaned_lines[i].startswith(header):
        for index, elem in enumerate(cleaned_lines[i + 1: i + 4]):
            if elem.startswith(header):
                next_index = i + 1 + index
                del cleaned_lines[i:next_index]
                i -= 1
    i += 1

grouped_lines = [cleaned_lines[n: n + 4] for n in range(0, len(cleaned_lines), 4)]

trimmed_dict: dict = {"diseases": [], "relationships": [], "cellular processes": []}
for group in grouped_lines:
    group.pop(0)

for group in grouped_lines:
    subject = group[0].split(":", 1)[1].strip()
    relation = group[1].split(":", 1)[1].strip()
    object = group[2].split(":", 1)[1].strip()
    trimmed_dict["diseases"].append(subject)
    trimmed_dict["relationships"].append(relation)
    trimmed_dict["cellular processes"].append(object)

for _key, value in trimmed_dict.copy().items():
    for i, elem in enumerate(value):
        if elem.lower() in (val.lower() for val in NULL_VALS):
            for _key, value in trimmed_dict.items():
                del value[i]

subjects_with_mondo = []
key = next(iter(trimmed_dict))
curr: list[str] = []
for i in range(len(trimmed_dict[key])):
    curr = []
    for value in trimmed_dict.values():
        curr.append(value[i])
    subjects_with_mondo.append(curr)
subjects_with_mondo = [x for x in subjects_with_mondo if x[0].startswith("MONDO:")]
mondo_go = [x for x in subjects_with_mondo if x[2].startswith("GO:")]

forprint(mondo_go)

subjects_with_names = []
# for elem in subjects_with_chebi:
for elem in subjects_with_mondo:  # type: ignore
    curr.append(elem)
    curr[0] = adapter.label(curr[0])
    subjects_with_names.append(curr)



from __future__ import annotations 

import re
import sys
from datetime import (
    date,
    datetime,
    time
)
from decimal import Decimal 
from enum import Enum 
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Literal,
    Optional,
    Union
)

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
    field_validator
)


metamodel_version = "None"
version = "None"


class ConfiguredBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment = True,
        validate_default = True,
        extra = "forbid",
        arbitrary_types_allowed = True,
        use_enum_values = True,
        strict = False,
    )
    pass




class LinkMLMeta(RootModel):
    root: Dict[str, Any] = {}
    model_config = ConfigDict(frozen=True)

    def __getattr__(self, key:str):
        return getattr(self.root, key)

    def __getitem__(self, key:str):
        return self.root[key]

    def __setitem__(self, key:str, value):
        self.root[key] = value

    def __contains__(self, key:str) -> bool:
        return key in self.root


linkml_meta = LinkMLMeta({'default_prefix': 'pp',
     'default_range': 'string',
     'description': 'A template for extracting information about predator-prey '
                    'relationships in ecological contexts',
     'id': 'https://w3id.org/ontogpt/predator_prey',
     'imports': ['linkml:types', 'core'],
     'license': 'https://creativecommons.org/publicdomain/zero/1.0/',
     'name': 'predator-prey-template',
     'prefixes': {'NCBITaxon': {'prefix_prefix': 'NCBITaxon',
                                'prefix_reference': 'http://purl.obolibrary.org/obo/NCBITAXON_'},
                  'RO': {'prefix_prefix': 'RO',
                         'prefix_reference': 'http://purl.obolibrary.org/obo/RO_'},
                  'linkml': {'prefix_prefix': 'linkml',
                             'prefix_reference': 'https://w3id.org/linkml/'},
                  'pp': {'prefix_prefix': 'pp',
                         'prefix_reference': 'http://w3id.org/ontogpt/predator-prey-template'},
                  'rdf': {'prefix_prefix': 'rdf',
                          'prefix_reference': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'}},
     'source_file': 'src/ontogpt/templates/predator_prey.yaml',
     'title': 'Predator Prey Relationship Template'} )

class NullDataOptions(str, Enum):
    UNSPECIFIED_METHOD_OF_ADMINISTRATION = "UNSPECIFIED_METHOD_OF_ADMINISTRATION"
    NOT_APPLICABLE = "NOT_APPLICABLE"
    NOT_MENTIONED = "NOT_MENTIONED"


class DietType(str, Enum):
    # An animal that primarily eats meat
    CARNIVORE = "CARNIVORE"
    # An animal that primarily eats plant material
    HERBIVORE = "HERBIVORE"
    # An animal that eats both meat and plant material
    OMNIVORE = "OMNIVORE"
    # An animal that primarily eats insects
    INSECTIVORE = "INSECTIVORE"
    # An animal that primarily eats fish
    PISCIVORE = "PISCIVORE"



class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    input_id: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'input_id', 'domain_of': ['ExtractionResult']} })
    input_title: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'input_title', 'domain_of': ['ExtractionResult']} })
    input_text: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'input_text', 'domain_of': ['ExtractionResult']} })
    raw_completion_output: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'raw_completion_output', 'domain_of': ['ExtractionResult']} })
    prompt: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'prompt', 'domain_of': ['ExtractionResult']} })
    extracted_object: Optional[Any] = Field(default=None, description="""The complex objects extracted from the text""", json_schema_extra = { "linkml_meta": {'alias': 'extracted_object', 'domain_of': ['ExtractionResult']} })
    named_entities: Optional[List[Any]] = Field(default=None, description="""Named entities extracted from the text""", json_schema_extra = { "linkml_meta": {'alias': 'named_entities', 'domain_of': ['ExtractionResult']} })


class NamedEntity(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'abstract': True, 'from_schema': 'http://w3id.org/ontogpt/core'})

    id: str = Field(default=..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(default=None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'annotations': {'owl': {'tag': 'owl',
                                 'value': 'AnnotationProperty, AnnotationAssertion'}},
         'domain_of': ['NamedEntity', 'Animal', 'PredationRelationshipType'],
         'slot_uri': 'rdfs:label'} })
    original_spans: Optional[List[str]] = Field(default=None, description="""The coordinates of the original text span from which the named entity was extracted, inclusive. For example, \"10:25\" means the span starting from the 10th character and ending with the 25th character. The first character in the text has index 0. Newlines are treated as single characters. Multivalued as there may be multiple spans for a single text.""", json_schema_extra = { "linkml_meta": {'alias': 'original_spans',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['This is determined during grounding and normalization',
                      'But is based on the full input text'],
         'domain_of': ['NamedEntity']} })

    @field_validator('original_spans')
    def pattern_original_spans(cls, v):
        pattern=re.compile(r"^\d+:\d+$")
        if isinstance(v,list):
            for element in v:
                if isinstance(v, str) and not pattern.match(element):
                    raise ValueError(f"Invalid original_spans format: {element}")
        elif isinstance(v,str):
            if not pattern.match(v):
                raise ValueError(f"Invalid original_spans format: {v}")
        return v


class CompoundExpression(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'abstract': True, 'from_schema': 'http://w3id.org/ontogpt/core'})

    pass


class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'abstract': True, 'from_schema': 'http://w3id.org/ontogpt/core'})

    subject: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'subject', 'domain_of': ['Triple']} })
    predicate: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'predicate', 'domain_of': ['Triple']} })
    object: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'object', 'domain_of': ['Triple']} })
    qualifier: Optional[str] = Field(default=None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""", json_schema_extra = { "linkml_meta": {'alias': 'qualifier', 'domain_of': ['Triple']} })
    subject_qualifier: Optional[str] = Field(default=None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""", json_schema_extra = { "linkml_meta": {'alias': 'subject_qualifier', 'domain_of': ['Triple']} })
    object_qualifier: Optional[str] = Field(default=None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""", json_schema_extra = { "linkml_meta": {'alias': 'object_qualifier', 'domain_of': ['Triple']} })


class TextWithTriples(ConfiguredBaseModel):
    """
    A text containing one or more relations of the Triple type.
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    publication: Optional[Publication] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'publication',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'domain_of': ['TextWithTriples', 'TextWithEntity']} })
    triples: Optional[List[Triple]] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'triples', 'domain_of': ['TextWithTriples']} })


class TextWithEntity(ConfiguredBaseModel):
    """
    A text containing one or more instances of a single type of entity.
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    publication: Optional[Publication] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'publication',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'domain_of': ['TextWithTriples', 'TextWithEntity']} })
    entities: Optional[List[str]] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'entities', 'domain_of': ['TextWithEntity']} })


class RelationshipType(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core',
         'id_prefixes': ['RO', 'biolink']})

    id: str = Field(default=..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(default=None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'annotations': {'owl': {'tag': 'owl',
                                 'value': 'AnnotationProperty, AnnotationAssertion'}},
         'domain_of': ['NamedEntity', 'Animal', 'PredationRelationshipType'],
         'slot_uri': 'rdfs:label'} })
    original_spans: Optional[List[str]] = Field(default=None, description="""The coordinates of the original text span from which the named entity was extracted, inclusive. For example, \"10:25\" means the span starting from the 10th character and ending with the 25th character. The first character in the text has index 0. Newlines are treated as single characters. Multivalued as there may be multiple spans for a single text.""", json_schema_extra = { "linkml_meta": {'alias': 'original_spans',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['This is determined during grounding and normalization',
                      'But is based on the full input text'],
         'domain_of': ['NamedEntity']} })

    @field_validator('original_spans')
    def pattern_original_spans(cls, v):
        pattern=re.compile(r"^\d+:\d+$")
        if isinstance(v,list):
            for element in v:
                if isinstance(v, str) and not pattern.match(element):
                    raise ValueError(f"Invalid original_spans format: {element}")
        elif isinstance(v,str):
            if not pattern.match(v):
                raise ValueError(f"Invalid original_spans format: {v}")
        return v


class Publication(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    id: Optional[str] = Field(default=None, description="""The publication identifier""", json_schema_extra = { "linkml_meta": {'alias': 'id', 'domain_of': ['NamedEntity', 'Publication']} })
    title: Optional[str] = Field(default=None, description="""The title of the publication""", json_schema_extra = { "linkml_meta": {'alias': 'title', 'domain_of': ['Publication']} })
    abstract: Optional[str] = Field(default=None, description="""The abstract of the publication""", json_schema_extra = { "linkml_meta": {'alias': 'abstract', 'domain_of': ['Publication']} })
    combined_text: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'combined_text', 'domain_of': ['Publication']} })
    full_text: Optional[str] = Field(default=None, description="""The full text of the publication""", json_schema_extra = { "linkml_meta": {'alias': 'full_text', 'domain_of': ['Publication']} })


class AnnotatorResult(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    subject_text: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'subject_text', 'domain_of': ['AnnotatorResult']} })
    object_id: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'object_id', 'domain_of': ['AnnotatorResult']} })
    object_text: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'object_text', 'domain_of': ['AnnotatorResult']} })


class PredatorPreyRelationship(Triple):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'https://w3id.org/ontogpt/predator_prey', 'tree_root': True})

    predator: Optional[str] = Field(default=None, description="""the animal that hunts and consumes other animals""", json_schema_extra = { "linkml_meta": {'alias': 'predator', 'domain_of': ['PredatorPreyRelationship']} })
    prey: Optional[str] = Field(default=None, description="""the animal that is hunted and consumed by other animals""", json_schema_extra = { "linkml_meta": {'alias': 'prey', 'domain_of': ['PredatorPreyRelationship']} })
    relationship_type: Optional[str] = Field(default=None, description="""the type of predatory relationship (e.g., obligate predation, opportunistic predation, scavenging)""", json_schema_extra = { "linkml_meta": {'alias': 'relationship_type', 'domain_of': ['PredatorPreyRelationship']} })
    habitat_context: Optional[List[str]] = Field(default=None, description="""the environment or habitat where this predator-prey interaction typically occurs""", json_schema_extra = { "linkml_meta": {'alias': 'habitat_context',
         'annotations': {'prompt': {'tag': 'prompt',
                                    'value': 'List the environments or habitats where '
                                             'this predator-prey interaction typically '
                                             'occurs. If there are multiple habitats, '
                                             'provide them as a semicolon-delimited '
                                             'list.'}},
         'domain_of': ['PredatorPreyRelationship']} })
    seasonal_pattern: Optional[str] = Field(default=None, description="""any seasonal patterns in the predator-prey relationship""", json_schema_extra = { "linkml_meta": {'alias': 'seasonal_pattern', 'domain_of': ['PredatorPreyRelationship']} })
    subject: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'subject', 'domain_of': ['Triple']} })
    predicate: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'predicate', 'domain_of': ['Triple']} })
    object: Optional[str] = Field(default=None, json_schema_extra = { "linkml_meta": {'alias': 'object', 'domain_of': ['Triple']} })
    qualifier: Optional[str] = Field(default=None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""", json_schema_extra = { "linkml_meta": {'alias': 'qualifier', 'domain_of': ['Triple']} })
    subject_qualifier: Optional[str] = Field(default=None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""", json_schema_extra = { "linkml_meta": {'alias': 'subject_qualifier', 'domain_of': ['Triple']} })
    object_qualifier: Optional[str] = Field(default=None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""", json_schema_extra = { "linkml_meta": {'alias': 'object_qualifier', 'domain_of': ['Triple']} })


class Animal(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'https://w3id.org/ontogpt/predator_prey'})

    label: Optional[str] = Field(default=None, description="""common name of the animal""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'domain_of': ['NamedEntity', 'Animal', 'PredationRelationshipType']} })
    scientific_name: Optional[str] = Field(default=None, description="""the scientific (Latin) name of the animal""", json_schema_extra = { "linkml_meta": {'alias': 'scientific_name', 'domain_of': ['Animal']} })
    taxonomic_rank: Optional[str] = Field(default=None, description="""the taxonomic classification of the animal (e.g., species, genus, family)""", json_schema_extra = { "linkml_meta": {'alias': 'taxonomic_rank', 'domain_of': ['Animal']} })
    diet_type: Optional[List[DietType]] = Field(default=None, description="""the type of diet of the animal (e.g., carnivore, herbivore, omnivore)""", json_schema_extra = { "linkml_meta": {'alias': 'diet_type',
         'annotations': {'prompt': {'tag': 'prompt',
                                    'value': 'The type of diet of the animal. This '
                                             'value must be one of: CARNIVORE,  '
                                             'HERBIVORE, OMNIVORE, INSECTIVORE, or '
                                             'PISCIVORE. If the animal has multiple  '
                                             'feeding strategies, provide them as a '
                                             'semicolon-delimited list (e.g., '
                                             '"CARNIVORE;PISCIVORE").'}},
         'domain_of': ['Animal']} })
    id: str = Field(default=..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    original_spans: Optional[List[str]] = Field(default=None, description="""The coordinates of the original text span from which the named entity was extracted, inclusive. For example, \"10:25\" means the span starting from the 10th character and ending with the 25th character. The first character in the text has index 0. Newlines are treated as single characters. Multivalued as there may be multiple spans for a single text.""", json_schema_extra = { "linkml_meta": {'alias': 'original_spans',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['This is determined during grounding and normalization',
                      'But is based on the full input text'],
         'domain_of': ['NamedEntity']} })

    @field_validator('original_spans')
    def pattern_original_spans(cls, v):
        pattern=re.compile(r"^\d+:\d+$")
        if isinstance(v,list):
            for element in v:
                if isinstance(v, str) and not pattern.match(element):
                    raise ValueError(f"Invalid original_spans format: {element}")
        elif isinstance(v,str):
            if not pattern.match(v):
                raise ValueError(f"Invalid original_spans format: {v}")
        return v


class PredationRelationshipType(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'https://w3id.org/ontogpt/predator_prey'})

    label: Optional[str] = Field(default=None, description="""the name of the predation relationship type""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'domain_of': ['NamedEntity', 'Animal', 'PredationRelationshipType']} })
    description: Optional[str] = Field(default=None, description="""a textual description of the predation relationship type""", json_schema_extra = { "linkml_meta": {'alias': 'description', 'domain_of': ['PredationRelationshipType']} })
    id: str = Field(default=..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    original_spans: Optional[List[str]] = Field(default=None, description="""The coordinates of the original text span from which the named entity was extracted, inclusive. For example, \"10:25\" means the span starting from the 10th character and ending with the 25th character. The first character in the text has index 0. Newlines are treated as single characters. Multivalued as there may be multiple spans for a single text.""", json_schema_extra = { "linkml_meta": {'alias': 'original_spans',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['This is determined during grounding and normalization',
                      'But is based on the full input text'],
         'domain_of': ['NamedEntity']} })

    @field_validator('original_spans')
    def pattern_original_spans(cls, v):
        pattern=re.compile(r"^\d+:\d+$")
        if isinstance(v,list):
            for element in v:
                if isinstance(v, str) and not pattern.match(element):
                    raise ValueError(f"Invalid original_spans format: {element}")
        elif isinstance(v,str):
            if not pattern.match(v):
                raise ValueError(f"Invalid original_spans format: {v}")
        return v


class EcologicalDocument(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'https://w3id.org/ontogpt/predator_prey'})

    predator_prey_relationships: Optional[List[PredatorPreyRelationship]] = Field(default=None, description="""semicolon separated list of predator-prey relationships""", json_schema_extra = { "linkml_meta": {'alias': 'predator_prey_relationships',
         'annotations': {'prompt': {'tag': 'prompt',
                                    'value': 'List all predator-prey relationships '
                                             'described in the document. If there are '
                                             'multiple  relationships, provide them as '
                                             'a semicolon-delimited list.'}},
         'domain_of': ['EcologicalDocument']} })
    animals: Optional[List[str]] = Field(default=None, description="""semicolon separated list of animals mentioned""", json_schema_extra = { "linkml_meta": {'alias': 'animals',
         'annotations': {'prompt': {'tag': 'prompt',
                                    'value': 'List all animals mentioned in the '
                                             'document. If there are multiple animals, '
                                             'provide them as a semicolon-delimited '
                                             'list.'}},
         'domain_of': ['EcologicalDocument']} })


class TaxonomicRank(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'annotations': {'annotators': {'tag': 'annotators',
                                        'value': 'sqlite:obo:ncbitaxon'}},
         'from_schema': 'https://w3id.org/ontogpt/predator_prey',
         'id_prefixes': ['NCBITaxon']})

    id: str = Field(default=..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(default=None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'annotations': {'owl': {'tag': 'owl',
                                 'value': 'AnnotationProperty, AnnotationAssertion'}},
         'domain_of': ['NamedEntity', 'Animal', 'PredationRelationshipType'],
         'slot_uri': 'rdfs:label'} })
    original_spans: Optional[List[str]] = Field(default=None, description="""The coordinates of the original text span from which the named entity was extracted, inclusive. For example, \"10:25\" means the span starting from the 10th character and ending with the 25th character. The first character in the text has index 0. Newlines are treated as single characters. Multivalued as there may be multiple spans for a single text.""", json_schema_extra = { "linkml_meta": {'alias': 'original_spans',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['This is determined during grounding and normalization',
                      'But is based on the full input text'],
         'domain_of': ['NamedEntity']} })

    @field_validator('original_spans')
    def pattern_original_spans(cls, v):
        pattern=re.compile(r"^\d+:\d+$")
        if isinstance(v,list):
            for element in v:
                if isinstance(v, str) and not pattern.match(element):
                    raise ValueError(f"Invalid original_spans format: {element}")
        elif isinstance(v,str):
            if not pattern.match(v):
                raise ValueError(f"Invalid original_spans format: {v}")
        return v


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
ExtractionResult.model_rebuild()
NamedEntity.model_rebuild()
CompoundExpression.model_rebuild()
Triple.model_rebuild()
TextWithTriples.model_rebuild()
TextWithEntity.model_rebuild()
RelationshipType.model_rebuild()
Publication.model_rebuild()
AnnotatorResult.model_rebuild()
PredatorPreyRelationship.model_rebuild()
Animal.model_rebuild()
PredationRelationshipType.model_rebuild()
EcologicalDocument.model_rebuild()
TaxonomicRank.model_rebuild()




"""Generic exporter class."""

from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import Optional, TextIO, Union

from linkml_runtime import SchemaView

from ontogpt.templates.core import ExtractionResult


def is_curie(s: str) -> bool:
    return ":" in s and " " not in s


@dataclass
class Exporter:
    def export(
        self,
        extraction_output: ExtractionResult,
        output: Union[str, Path, TextIO, BytesIO],
        schemaview: Optional[SchemaView],
    ):
        raise NotImplementedError



"""YAML Wrapper."""

import io
import logging
from typing import Any, Optional, TextIO

from ruamel.yaml import YAML, RoundTripRepresenter

from ontogpt.io.utils import eliminate_empty

logger = logging.getLogger(__name__)


def repr_str(dumper: RoundTripRepresenter, data: str):
    if "\n" in data:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


def dump_minimal_yaml(obj: Any, minimize=True, file: Optional[TextIO] = None) -> str:
    """Dump a YAML string, but eliminate Nones and empty lists and dicts."""
    yaml = YAML()
    yaml.representer.add_representer(str, repr_str)
    yaml.default_flow_style = False
    yaml.indent(sequence=4, offset=2)
    if not file:
        file = io.StringIO()
        yaml.dump(eliminate_empty(obj, not minimize), file)
        return file.getvalue()
    else:
        yaml.dump(eliminate_empty(obj, not minimize), file)
        return ""



"""Client for running LLM completion requests through LiteLLM."""

import logging
import sys
from dataclasses import dataclass, field

import litellm
import numpy as np
import openai  # For error handling
from litellm import completion, embedding
from litellm.caching.caching import Cache
from oaklib.utilities.apikey_manager import get_apikey_value

from ontogpt import DEFAULT_MODEL, MODELS

logger = logging.getLogger(__name__)

# Just get the part before the slash in each model name
SERVICES = {model.split("/")[0] for model in MODELS.keys() if len(model.split("/")) > 1}

# Necessary to avoid repeated debug messages
litellm.suppress_debug_info = True


@dataclass
class LLMClient:

    model: str = field(default_factory=lambda: DEFAULT_MODEL)
    cache_db_path: str = ""
    api_key: str = ""
    api_base: str = None
    api_version: str = None

    # litellm uses this param to specify client when it isn't clear from
    # the model name
    custom_llm_provider: str = None

    temperature: float = 1.0

    system_message: str = ""
    """System message to be provided to the LLM."""

    def __post_init__(self):
        # Get appropriate API key for the model source
        # and other details if needed.
        # This will look at any env vars FIRST,
        # then configs handled by oaklib.

        # Need to check on the validity of the model name first.
        # Check if the model name is a string first.
        # if not, try to make it one
        if not isinstance(self.model, str):
            if isinstance(self.model, (tuple, list)) and len(self.model) > 0:
                self.model = str(self.model[0])
                logger.warning(f"Model name was a {type(self.model).__name__}.")
                logger.warning(f"Converted to string: {self.model}")
            else:
                raise ValueError(f"Model name must be a string, got {type(self.model)}")

        if self.model.startswith("ollama"):
            self.api_key = ""  # Don't need an API key
        elif self.model.startswith("fake"):
            # Just used for testing
            self.api_key = ""  # Don't need an API key
            logger.info(f"Using mock model: {self.model}")
        elif not self.api_key and not self.custom_llm_provider:
            self.api_key = get_apikey_value("openai")
        elif self.custom_llm_provider == "anthropic":
            self.api_key = get_apikey_value("anthropic-key")
        elif self.custom_llm_provider == "groq":
            self.api_key = get_apikey_value("groq-key")
        else:
            for service in SERVICES:
                if self.model.startswith(service):
                    self.api_key = get_apikey_value(service + "-key")
                    if service == "azure" and self.api_base is None:
                        self.api_base = get_apikey_value(service + "-base")
                    if service == "azure" and self.api_version is None:
                        self.api_version = get_apikey_value(service + "-version")

        # Set up the cache, and set the cache path if provided
        if len(self.cache_db_path) == 0:
            litellm.cache = Cache(disk_cache_dir="./.litellm_cache")
        else:
            litellm.cache = Cache(disk_cache_dir=self.cache_db_path)

    def complete(self, prompt, show_prompt: bool = False, **kwargs) -> str:

        logger.info(f"Complete: engine={self.model}, prompt[{len(prompt)}]={prompt[0:100]}...")
        if show_prompt:
            logger.info(f" SENDING PROMPT:\n{prompt}")

        response = None

        these_messages = [{"content": prompt, "role": "user"}]

        if self.system_message:
            these_messages.insert(0, {"content": self.system_message, "role": "system"})

        # This toggle controls whether we can continue or not.
        # Some errors may be temporary, while others, such as authentication errors,
        # require action before we may continue.
        force_stop = False

        try:
            # TODO: expose user prompt to CLI
            response = completion(
                api_key=self.api_key,
                api_base=self.api_base,
                api_version=self.api_version,
                model=self.model,
                messages=these_messages,
                temperature=self.temperature,
                caching=True,
                custom_llm_provider=self.custom_llm_provider,
            )
        except openai.APITimeoutError as e:
            logger.error(f"Encountered API timeout error: {e}")
        except litellm.exceptions.AuthenticationError as e:
            logger.error(f"Encountered authentication error: {e}")
            force_stop = True
        except litellm.exceptions.NotFoundError as e:
            logger.error(f"Encountered error due to unrecognized model or endpoint: {e}")
            force_stop = True
        except litellm.exceptions.BadRequestError as e:
            logger.error(f"Encountered error due to bad request: {e}")
            force_stop = True
        except litellm.exceptions.UnprocessableEntityError as e:
            logger.error(f"Encountered error due to unprocessable entity: {e}")
        except litellm.exceptions.PermissionDeniedError as e:
            logger.error(f"Encountered error - permission denied: {e}")
            force_stop = True
        except litellm.exceptions.RateLimitError as e:
            logger.error(f"Encountered rate limiting: {e}")
        except litellm.exceptions.ContextWindowExceededError as e:
            logger.error(f"Exceeded context window: {e}")
        except litellm.exceptions.ServiceUnavailableError as e:
            logger.error(f"Service unavailable: {e}")
            force_stop = True
        except litellm.exceptions.InternalServerError as e:
            logger.error(f"Internal server error: {e}")
            force_stop = True
        except litellm.exceptions.APIError as e:
            logger.error(f"API returned an invalid response: {e}")
        except litellm.exceptions.APIConnectionError as e:
            logger.error(f"API connection error: {e}")
        except Exception as e:
            logger.error(f"Encountered error: {type(e)}, Error: {e}")

        if force_stop:
            sys.exit("Exiting...")

        if response is not None:
            payload = response.choices[0].message.content
        else:
            logger.error("No response or response is empty.")
            payload = ""

        return payload

    def embeddings(self, text: str):
        text = str(text)

        # TODO: set embedding model based on model source
        # Or at least set the default for OpenAI models
        if self.model is None:
            model = "text-embedding-ada-002"
        else:
            model = self.model

        logger.info(f"Retrieving embeddings from {model} for text: {text[0:80]}...")

        response = embedding(
            api_key=self.api_key,
            api_base=self.api_base,
            api_version=self.api_version,
            model=model,
            input=[text],
            caching=True,
        )

        if response is not None:
            payload = response.data[0]["embedding"]
        else:
            logger.error("No response or response is empty.")
            payload = ""

        return payload

    def similarity(self, text1: str, text2: str, **kwargs):
        a1 = self.embeddings(text1, **kwargs)
        a2 = self.embeddings(text2, **kwargs)
        logger.debug(f"similarity: {a1[0:10]}... x {a2[0:10]}... // ({len(a1)} x {len(a2)})")
        return np.dot(a1, a2) / (np.linalg.norm(a1) * np.linalg.norm(a2))

    def euclidian_distance(self, text1: str, text2: str, **kwargs):
        a1 = self.embeddings(text1, **kwargs)
        a2 = self.embeddings(text2, **kwargs)
        return np.linalg.norm(np.array(a1) - np.array(a2))



"""HTML Exporter."""

import html
from dataclasses import dataclass
from io import BytesIO, StringIO, TextIOWrapper
from pathlib import Path
from typing import Any, Optional, TextIO, Union

import pydantic
import yaml
from linkml_runtime import SchemaView

from ontogpt.io.exporter import Exporter, is_curie
from ontogpt.templates.core import ExtractionResult


@dataclass
class HTMLExporter(Exporter):
    """
    An exporter that will generate HTML for extraction results.

    TODO: rewrite to use bootstrap
    """

    output: Optional[Union[BytesIO, TextIO]]

    def export(
        self,
        extraction_output: ExtractionResult,
        output: Union[str, Path, TextIO, BytesIO],
        schemaview: Optional[SchemaView] = None,
    ):  # type: ignore
        if isinstance(output, Path):
            output = open(str(output), "w", encoding="utf-8")
        if isinstance(output, str):
            output = StringIO(output)
        self.output = output
        self.export_metadata(extraction_output)
        self.export_results(extraction_output.extracted_object, extraction_output)
        self.h2("YAML Object")
        self.details(yaml.dump(extraction_output.model_dump()), output, code="yaml")
        self.h2("Prompt")
        self.details(extraction_output.prompt, output)
        self.h2("Completion")
        self.details(extraction_output.raw_completion_output, output)

    def export_metadata(self, extraction_output: ExtractionResult):
        output = self.output
        self.h1("Extraction Results:\n\n")
        self.h2("Input")
        self.i(extraction_output.input_text)

    def export_results(self, obj: Any, extraction_output: ExtractionResult):
        self.h2("Results")
        self.export_object(obj, extraction_output, -1)

    def export_object(
        self, obj: pydantic.BaseModel, extraction_output: ExtractionResult, indent: int
    ):
        self.open_div()
        if indent >= 0:
            self.open_ul()
        # There may not be an object returned due to errors.
        if obj is None:
            self.h3("<i>Encountered an error. See raw completion output for details.</i>")
        else:
            for field_name, _field in obj.model_fields.items():
                if indent < 0:
                    self.h3(field_name)
                else:
                    self.li(f"<i>{field_name}</i>: ")
                value = getattr(obj, field_name)
                if isinstance(value, pydantic.BaseModel):
                    self.export_object(value, extraction_output, indent + 1)
                elif isinstance(value, list):
                    self.open_ul()
                    n = 1
                    for item in value:
                        self.li(f"<b>item: {n}</b>: ")
                        n += 1
                        if isinstance(item, pydantic.BaseModel):
                            self.export_object(
                                item, extraction_output=extraction_output, indent=indent + 1
                            )
                        else:
                            self.export_atom(item, extraction_output, indent + 1)
                    self.close_ul()
                else:
                    self.export_atom(value, extraction_output, indent + 1)
        if indent >= 0:
            self.open_ul()
        self.close_div()

    def export_atom(self, value, extraction_output: ExtractionResult, indent: int):
        output = self.output
        matches = [
            ne for ne in extraction_output.named_entities if ne.id == value and is_curie(ne.id)
        ]
        if isinstance(output, BytesIO):
            output = TextIOWrapper(output, encoding="utf-8")
        if matches:
            match = matches[0]
            output.write(f"{match.label} {self.link(match.id)}")
        else:
            output.write(str(value))
        output.write("\n")

    def details(self, text: Optional[str], output: Union[BytesIO, TextIO], code: str = ""):
        if isinstance(output, BytesIO):
            output = TextIOWrapper(output, encoding="utf-8")
        output.write("<details>\n")
        output.write("<pre>\n")
        self.w(text)
        output.write("\n</pre>\n")
        output.write("\n</details>\n")

    def link(self, curie: str) -> str:
        return f'<a href="https://bioregistry.io/{curie}">{curie}</a>'

    def open_ul(self):
        self.output.write("<ul>\n")

    def close_ul(self):
        self.output.write("</ul>\n")

    def open_div(self):
        self.output.write("<div>\n")

    def close_div(self):
        self.output.write("</div>\n")

    def li(self, text: str = ""):
        self.tag("li", text)

    def h1(self, text: str):
        self.tag("h1", text)

    def h2(self, text: str):
        self.tag("h1", text)

    def h3(self, text: str):
        self.tag("h3", text)

    def i(self, text: Optional[str]):
        self.tag("i", html.escape(text))

    def tag(self, tag: str, text: str):
        if isinstance(self.output, BytesIO):
            self.output = TextIOWrapper(self.output, encoding="utf-8")
        self.output.write(f"<{tag}>{text}</{tag}>\n")

    def w(self, text: str):
        if isinstance(self.output, BytesIO):
            self.output = TextIOWrapper(self.output, encoding="utf-8")
        self.output.write(html.escape(text))



"""HPOA evaluation."""
import csv
import logging
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from random import shuffle
from typing import Dict, Iterable, Iterator, List, Optional, Tuple

from oaklib import BasicOntologyInterface, get_adapter
from oaklib.datamodels.search import SearchConfiguration
from oaklib.datamodels.search_datamodel import SearchProperty
from oaklib.interfaces import SearchInterface
from pydantic import BaseModel

from ontogpt.clients.pubmed_client import PubmedClient
from ontogpt.engines.spires_engine import SPIRESEngine
from ontogpt.evaluation.evaluation_engine import SimilarityScore, SPIRESEvaluationEngine
from ontogpt.templates.mendelian_disease import MendelianDisease

DATABASE_DIR = Path(__file__).parent / "database"
TEST_CASES_DIR = Path("tests").joinpath("input")
TEST_HPOA_FILE = "test_sample.hpoa.tsv"
NUM_TESTS = 3  # Note: each test requires input text; see provided test cases

DISEASE_ID = str
TERM = str
ASPECT = str
PUBLICATION = str


class PredictionHPOA(BaseModel):
    predicted_object: Optional[MendelianDisease] = None
    test_object: Optional[MendelianDisease] = None
    scores: Optional[Dict[str, SimilarityScore]] = None

    def calculate_scores(self):
        self.scores = {}
        for k in ["synonyms", "subclass_of", "symptoms", "disease_onsets", "genes"]:
            self.scores[k] = SimilarityScore.from_set(
                getattr(self.test_object, k, []),
                getattr(self.predicted_object, k, []),
            )
        for k in ["description", "name", "inheritance"]:
            if getattr(self.test_object, k):
                self.scores[k] = SimilarityScore.from_set(
                    getattr(self.test_object, k, "").split(),
                    getattr(self.predicted_object, k, "").split(),
                )


class EvaluationObjectSetHPOA(BaseModel):
    """A result of predicting paths."""

    test: List[MendelianDisease] = None
    training: List[MendelianDisease] = None
    predictions: List = []


class HPOAnnotation(BaseModel):
    subject: DISEASE_ID = ""
    term: TERM = ""
    aspect: ASPECT = ""
    publication: PUBLICATION = ""


@dataclass
class EvalHPOA(SPIRESEvaluationEngine):
    mondo: BasicOntologyInterface = None

    def __post_init__(self):
        self.extractor = SPIRESEngine("mendelian_disease.MendelianDisease")
        self.mondo = get_adapter("sqlite:obo:mondo")

    def load_test_cases(self) -> List[MendelianDisease]:
        return []

    def disease_text(self, id: str):
        id = id.lower().replace(":", "-")
        with open(TEST_CASES_DIR / "cases" / f"{id}.txt") as f:
            return f.read()

    def parse_hpoa(self) -> Iterator[HPOAnnotation]:
        with open(TEST_CASES_DIR / TEST_HPOA_FILE) as file:
            reader = csv.reader(file, delimiter="\t")
            for row in reader:
                yield HPOAnnotation(
                    subject=row[0],
                    term=row[3],
                    publication=row[4],
                    aspect=row[10],
                )

    def annotations_to_diseases(
        self, anns: Iterable[HPOAnnotation] = None
    ) -> List[MendelianDisease]:
        if anns is None:
            anns = self.parse_hpoa()
        diseases: Dict[str, MendelianDisease] = {}
        for ann in anns:
            subject = ann.subject
            term = ann.term
            pub = ann.publication
            aspect = ann.aspect
            if subject not in diseases:
                diseases[subject] = MendelianDisease(id=subject)
            disease = diseases[subject]
            # print(f"Adding {term} to {subject} in {aspect}")
            if aspect == "P" and disease.symptoms is not None:
                disease.symptoms.append(term)
            elif aspect == "C" and disease.disease_onsets is not None:
                disease.disease_onsets.append(term)
            elif aspect == "I":
                disease.inheritance = term
            if pub and pub not in disease.publications and disease.publications is not None:
                disease.publications.append(pub)
        return list(diseases.values())

    def diseases(self, anns: Iterable[HPOAnnotation] = None) -> List[MendelianDisease]:
        diseases = self.annotations_to_diseases(anns)
        for disease in diseases:
            self.enhance(disease)
        return diseases

    def diseases_by_publication(self) -> Dict[Tuple[DISEASE_ID, PUBLICATION], MendelianDisease]:
        anns = list(self.parse_hpoa())
        tuple2ann = defaultdict(list)
        for ann in anns:
            if ann.publication and ann.publication.startswith("PMID:"):
                tuple2ann[(ann.subject, ann.publication)].append(ann)
        tuple2disease = {}
        for k, anns in tuple2ann.items():
            diseases = self.annotations_to_diseases(anns)
            if len(diseases) != 1:
                raise AssertionError(f"Expected 1 disease, got {len(diseases)} for {k}")
            disease = diseases[0]
            self.enhance(disease)
            tuple2disease[k] = disease
        return tuple2disease

    def enhance(self, obj: MendelianDisease):
        mondo = self.mondo
        config = SearchConfiguration(properties=[SearchProperty.MAPPED_IDENTIFIER])
        if not isinstance(mondo, SearchInterface):
            raise ValueError("Mondo is not a SearchInterface")
        entities = list(mondo.basic_search(obj.id, config))
        if len(entities) == 0:
            logging.warning(f"No matches for {obj.id} => {entities}")
            return
        if len(entities) > 1:
            logging.warning(f"multiple matches for {obj.id} => {entities}")
        entity = entities[0]
        obj.name = mondo.label(entity)
        obj.label = obj.name
        obj.description = mondo.definition(entity)
        obj.subclass_of = list(mondo.hierarchical_parents(entity))
        obj.synonyms = list(mondo.entity_aliases(entity))
        for _s, _p, gene in mondo.relationships([entity], ["RO:0004003"]):
            gene = (
                gene.replace("http://identifiers.org/hgnc/", "HGNC:")
                .replace("<", "")
                .replace(">", "")
            )
            obj.genes.append(gene)
        return obj

    def eval(self, task: str = "", **kwargs) -> EvaluationObjectSetHPOA:
        if not task or task == "pubs":
            return self.eval_against_pubs(**kwargs)
        elif task == "omim":
            return self.eval_against_omim(**kwargs)
        elif task == "all":
            return self.eval_against_omim_plus_pubs(**kwargs)
        else:
            raise ValueError(f"Unknown task {task}")

    def eval_against_pubs(self, num_tests=NUM_TESTS) -> EvaluationObjectSetHPOA:
        ke = self.extractor
        pmc = PubmedClient()
        eos = EvaluationObjectSetHPOA()
        eos.test = list(self.diseases_by_publication().values())
        eos.training = []
        eos.predictions = []
        shuffle(eos.test)
        for test_case in eos.test[0 : num_tests - 1]:
            # text = self.disease_text(test_case.id)
            if len(test_case.publications) != 1:
                raise ValueError(f"Expected 1 publication, got {len(test_case.publications)}")
            pub = test_case.publications[0]
            text = pmc.text(pub)
            if type(text) is list:  # Shouldn't happen, but could happen
                for entry in text:
                    results = ke.extract_from_text(entry)
            elif type(text) is str:
                results = ke.extract_from_text(text)
            predicted_obj = results.extracted_object
            pred = PredictionHPOA(predicted_object=predicted_obj, test_object=test_case)
            pred.calculate_scores()
            eos.predictions.append(pred)
        return eos

    def eval_against_omim(self, **kwargs) -> EvaluationObjectSetHPOA:
        return self.eval_against_omim_or_pubs(use_publications=False)

    def eval_against_omim_plus_pubs(self, **kwargs) -> EvaluationObjectSetHPOA:
        return self.eval_against_omim_or_pubs(use_publications=True)

    def eval_against_omim_or_pubs(
        self, num_tests=NUM_TESTS, use_publications=False
    ) -> EvaluationObjectSetHPOA:
        ke = self.extractor
        eos = EvaluationObjectSetHPOA()
        eos.test = self.diseases()
        eos.training = []
        eos.predictions = []
        shuffle(eos.test)
        for test_case in eos.test[0:num_tests]:
            text = self.disease_text(test_case.id)
            stub = {"name": test_case.name}
            results = ke.extract_from_text(text, object=stub)
            if use_publications:
                pmc = PubmedClient()
                pub_resultset = []
                for pmid in test_case.publications:
                    if str(pmid).startswith("PMID:"):
                        pub_text = pmc.text(pmid)
                        if type(pub_text) is str:  # Should just be one document.
                            pub_resultset.append(ke.extract_from_text(pub_text, object=stub))
                        else:
                            continue
                results = ke.merge_resultsets([results] + pub_resultset, ["name"])
            predicted_obj = results.extracted_object
            pred = PredictionHPOA(predicted_object=predicted_obj, test_object=test_case)
            pred.calculate_scores()
            eos.predictions.append(pred)
        return eos



"""Utilities for parsing text."""

import re

from typing import List


def split_on_one_of(text: str, separators: List[str]) -> List[str]:
    """Split text on the first separator found."""
    for sep in separators:
        if sep in text:
            return text.split(sep)
    return [text]


def get_section_of_interest(data, tag_of_interest):
    # I blame adobe
    # Find the index of the element that matches the case-insensitive regex pattern
    start_index = None
    pattern = re.compile(tag_of_interest, re.IGNORECASE)
    if isinstance(data, str):
        data = data.split("\n")
    for i, item in enumerate(data):
        if pattern.search(item):
            start_index = i
            break

    if start_index is not None:
        # Find the index of the next element that starts with '<p>'
        next_index = next(
            (
                i
                for i, item in enumerate(data[start_index + 1:], start=start_index + 1)
                if item.startswith("<p>")
            ),
            None,
        )

        if next_index is not None:
            # Extract the desired element
            result = data[next_index]
            return result
        else:
            raise ValueError("No element starting with '<p>' found after the tag_of_interest")
    else:
        raise ValueError("No element matching the tag_of_interest found in the list.")


def get_span_values(text: str, find_text: str) -> List[str]:
    """
    Find spans in text.

    Given an input text and a text to find, return a list of span values
    containing the text to find. This is a list because the text may appear
    multiple times.
    The coordinates are inclusive. For example, "10:25" means the span starting
    from the 10th character and ending with the 25th character. The first
    character in the text has index 0.
    All newlines are treated as single characters.
    """
    span_values = []
    start = 0

    pattern = re.compile(re.escape(find_text), re.IGNORECASE)

    # Remove all newlines, replace with space
    text = text.replace("\n", " ")

    for match in pattern.finditer(text):
        start = match.start()
        end = match.end() - 1
        span_values.append(f"{start}:{end}")

    return span_values



from __future__ import annotations 
from datetime import (
    datetime,
    date
)
from decimal import Decimal 
from enum import Enum 
import re
import sys
from typing import (
    Any,
    ClassVar,
    List,
    Literal,
    Dict,
    Optional,
    Union
)
from pydantic.version import VERSION  as PYDANTIC_VERSION 
if int(PYDANTIC_VERSION[0])>=2:
    from pydantic import (
        BaseModel,
        ConfigDict,
        Field,
        RootModel,
        field_validator
    )
else:
    from pydantic import (
        BaseModel,
        Field,
        validator
    )

metamodel_version = "None"
version = "None"


class ConfiguredBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment = True,
        validate_default = True,
        extra = "forbid",
        arbitrary_types_allowed = True,
        use_enum_values = True,
        strict = False,
    )
    pass




class LinkMLMeta(RootModel):
    root: Dict[str, Any] = {}
    model_config = ConfigDict(frozen=True)

    def __getattr__(self, key:str):
        return getattr(self.root, key)

    def __getitem__(self, key:str):
        return self.root[key]

    def __setitem__(self, key:str, value):
        self.root[key] = value

    def __contains__(self, key:str) -> bool:
        return key in self.root


linkml_meta = LinkMLMeta({'default_prefix': 'miro',
     'default_range': 'string',
     'description': 'A template for extracting the minimal information for '
                    'reporting an ontology, as per the MIRO guidelines. See '
                    'doi:10.1186/s13326-017-0172-7 The target for this template '
                    'should be a report or publication describing a single '
                    'ontology.',
     'id': 'http://w3id.org/ontogpt/miro',
     'imports': ['linkml:types', 'core'],
     'license': 'https://creativecommons.org/publicdomain/zero/1.0/',
     'name': 'miro',
     'prefixes': {'linkml': {'prefix_prefix': 'linkml',
                             'prefix_reference': 'https://w3id.org/linkml/'},
                  'onto_usage': {'prefix_prefix': 'onto_usage',
                                 'prefix_reference': 'http://w3id.org/ontogpt/onto_usage'},
                  'rdf': {'prefix_prefix': 'rdf',
                          'prefix_reference': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'}},
     'source_file': 'src/ontogpt/templates/miro.yaml',
     'title': 'MIRO Extraction Template'} )

class NullDataOptions(str, Enum):
    UNSPECIFIED_METHOD_OF_ADMINISTRATION = "UNSPECIFIED_METHOD_OF_ADMINISTRATION"
    NOT_APPLICABLE = "NOT_APPLICABLE"
    NOT_MENTIONED = "NOT_MENTIONED"



class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    input_id: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'input_id', 'domain_of': ['ExtractionResult']} })
    input_title: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'input_title', 'domain_of': ['ExtractionResult']} })
    input_text: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'input_text', 'domain_of': ['ExtractionResult']} })
    raw_completion_output: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'raw_completion_output', 'domain_of': ['ExtractionResult']} })
    prompt: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'prompt', 'domain_of': ['ExtractionResult']} })
    extracted_object: Optional[Any] = Field(None, description="""The complex objects extracted from the text""", json_schema_extra = { "linkml_meta": {'alias': 'extracted_object', 'domain_of': ['ExtractionResult']} })
    named_entities: Optional[List[Any]] = Field(default_factory=list, description="""Named entities extracted from the text""", json_schema_extra = { "linkml_meta": {'alias': 'named_entities', 'domain_of': ['ExtractionResult']} })


class NamedEntity(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'abstract': True, 'from_schema': 'http://w3id.org/ontogpt/core'})

    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })


class CompoundExpression(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'abstract': True, 'from_schema': 'http://w3id.org/ontogpt/core'})

    pass


class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'abstract': True, 'from_schema': 'http://w3id.org/ontogpt/core'})

    subject: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'subject', 'domain_of': ['Triple']} })
    predicate: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'predicate', 'domain_of': ['Triple']} })
    object: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'object', 'domain_of': ['Triple']} })
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""", json_schema_extra = { "linkml_meta": {'alias': 'qualifier', 'domain_of': ['Triple']} })
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""", json_schema_extra = { "linkml_meta": {'alias': 'subject_qualifier', 'domain_of': ['Triple']} })
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""", json_schema_extra = { "linkml_meta": {'alias': 'object_qualifier', 'domain_of': ['Triple']} })


class TextWithTriples(ConfiguredBaseModel):
    """
    A text containing one or more relations of the Triple type.
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    publication: Optional[Publication] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'publication', 'domain_of': ['TextWithTriples', 'TextWithEntity']} })
    triples: Optional[List[Triple]] = Field(default_factory=list, json_schema_extra = { "linkml_meta": {'alias': 'triples', 'domain_of': ['TextWithTriples']} })


class TextWithEntity(ConfiguredBaseModel):
    """
    A text containing one or more instances of a single type of entity.
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    publication: Optional[Publication] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'publication', 'domain_of': ['TextWithTriples', 'TextWithEntity']} })
    entities: Optional[List[str]] = Field(default_factory=list, json_schema_extra = { "linkml_meta": {'alias': 'entities', 'domain_of': ['TextWithEntity']} })


class RelationshipType(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core',
         'id_prefixes': ['RO', 'biolink']})

    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })


class Publication(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    id: Optional[str] = Field(None, description="""The publication identifier""", json_schema_extra = { "linkml_meta": {'alias': 'id', 'domain_of': ['NamedEntity', 'Publication']} })
    title: Optional[str] = Field(None, description="""The title of the publication""", json_schema_extra = { "linkml_meta": {'alias': 'title', 'domain_of': ['Publication']} })
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""", json_schema_extra = { "linkml_meta": {'alias': 'abstract', 'domain_of': ['Publication']} })
    combined_text: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'combined_text', 'domain_of': ['Publication']} })
    full_text: Optional[str] = Field(None, description="""The full text of the publication""", json_schema_extra = { "linkml_meta": {'alias': 'full_text', 'domain_of': ['Publication']} })


class AnnotatorResult(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    subject_text: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'subject_text', 'domain_of': ['AnnotatorResult']} })
    object_id: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'object_id', 'domain_of': ['AnnotatorResult']} })
    object_text: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'object_text', 'domain_of': ['AnnotatorResult']} })


class Ontology(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/miro', 'tree_root': True})

    name: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'name', 'domain_of': ['Ontology', 'Owner', 'CitedOntology']} })
    owners: Optional[List[str]] = Field(default_factory=list, json_schema_extra = { "linkml_meta": {'alias': 'owners', 'domain_of': ['Ontology']} })
    license: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'license', 'domain_of': ['Ontology']} })
    url: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'url', 'domain_of': ['Ontology']} })
    repository: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'repository', 'domain_of': ['Ontology']} })
    methodological_framework: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'methodological_framework', 'domain_of': ['Ontology']} })
    need: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'need', 'domain_of': ['Ontology']} })
    competition: Optional[List[str]] = Field(default_factory=list, json_schema_extra = { "linkml_meta": {'alias': 'competition', 'domain_of': ['Ontology']} })
    target_audience: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'target_audience', 'domain_of': ['Ontology']} })
    scope_and_coverage: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'scope_and_coverage', 'domain_of': ['Ontology']} })
    development_community: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'development_community', 'domain_of': ['Ontology']} })
    communication: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'communication', 'domain_of': ['Ontology']} })
    knowledge_acquisition_method: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'knowledge_acquisition_method', 'domain_of': ['Ontology']} })
    source_knowledge_location: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'source_knowledge_location', 'domain_of': ['Ontology']} })
    content_selection: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'content_selection', 'domain_of': ['Ontology']} })
    knowledge_representation_language: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'knowledge_representation_language', 'domain_of': ['Ontology']} })
    development_environment: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'development_environment', 'domain_of': ['Ontology']} })
    ontology_metrics: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'ontology_metrics', 'domain_of': ['Ontology']} })
    incorporation_of_ontologies: Optional[List[str]] = Field(default_factory=list, json_schema_extra = { "linkml_meta": {'alias': 'incorporation_of_ontologies', 'domain_of': ['Ontology']} })
    entity_naming_convention: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'entity_naming_convention', 'domain_of': ['Ontology']} })
    identifier_generation_policy: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'identifier_generation_policy', 'domain_of': ['Ontology']} })
    entity_metadata_policy: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'entity_metadata_policy', 'domain_of': ['Ontology']} })
    upper_ontology: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'upper_ontology', 'domain_of': ['Ontology']} })
    ontology_relationships: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'ontology_relationships', 'domain_of': ['Ontology']} })
    axiom_patterns: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'axiom_patterns', 'domain_of': ['Ontology']} })
    dereferencable_iri: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'dereferencable_iri', 'domain_of': ['Ontology']} })
    sustainability_plan: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'sustainability_plan', 'domain_of': ['Ontology']} })
    deprecation_strategy: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'deprecation_strategy', 'domain_of': ['Ontology']} })
    versioning_policy: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'versioning_policy', 'domain_of': ['Ontology']} })
    testing: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'testing', 'domain_of': ['Ontology']} })
    evaluation: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'evaluation', 'domain_of': ['Ontology']} })
    examples_of_use: Optional[List[UseExample]] = Field(default_factory=list, json_schema_extra = { "linkml_meta": {'alias': 'examples_of_use', 'domain_of': ['Ontology']} })
    institutional_endorsement: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'institutional_endorsement', 'domain_of': ['Ontology']} })
    evidence_of_use: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'evidence_of_use', 'domain_of': ['Ontology']} })
    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })


class Owner(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'annotations': {'owl': {'tag': 'owl', 'value': 'IntersectionOf'}},
         'from_schema': 'http://w3id.org/ontogpt/miro'})

    name: Optional[str] = Field(None, description="""The name of a person, group of people, organization, or consortium that manages development of the ontology.""", json_schema_extra = { "linkml_meta": {'alias': 'name', 'domain_of': ['Ontology', 'Owner', 'CitedOntology']} })
    affiliation: Optional[str] = Field(None, description="""The name of the organization with which the owner is affiliated.""", json_schema_extra = { "linkml_meta": {'alias': 'affiliation', 'domain_of': ['Owner']} })
    contact: Optional[str] = Field(None, description="""The contact details of the owner. This is typically an email address.""", json_schema_extra = { "linkml_meta": {'alias': 'contact', 'domain_of': ['Owner']} })
    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })


class CitedOntology(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/miro'})

    name: Optional[str] = Field(None, description="""The name of the ontology being cited.""", json_schema_extra = { "linkml_meta": {'alias': 'name', 'domain_of': ['Ontology', 'Owner', 'CitedOntology']} })
    version: Optional[str] = Field(None, description="""The version of the ontology being cited.""", json_schema_extra = { "linkml_meta": {'alias': 'version', 'domain_of': ['CitedOntology']} })
    citation: Optional[str] = Field(None, description="""The citation for the ontology being cited.""", json_schema_extra = { "linkml_meta": {'alias': 'citation', 'domain_of': ['CitedOntology']} })
    integration: Optional[str] = Field(None, description="""A description of how the ontology being cited is integrated into the ontology being reported.""", json_schema_extra = { "linkml_meta": {'alias': 'integration', 'domain_of': ['CitedOntology']} })
    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })


class UseExample(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/miro'})

    example_description: Optional[str] = Field(None, description="""A description of the ontology in use in its application setting or use case.""", json_schema_extra = { "linkml_meta": {'alias': 'example_description', 'domain_of': ['UseExample']} })


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
ExtractionResult.model_rebuild()
NamedEntity.model_rebuild()
CompoundExpression.model_rebuild()
Triple.model_rebuild()
TextWithTriples.model_rebuild()
TextWithEntity.model_rebuild()
RelationshipType.model_rebuild()
Publication.model_rebuild()
AnnotatorResult.model_rebuild()
Ontology.model_rebuild()
Owner.model_rebuild()
CitedOntology.model_rebuild()
UseExample.model_rebuild()




"""BeautifulSoup client."""

from dataclasses import dataclass

import requests
from bs4 import BeautifulSoup


@dataclass
class SoupClient:
    """A client for the Beautiful Soup API."""

    def text(self, url: str) -> str:
        """Get the text of a web site from its HTML.

        :param url: string containing the URL to extract text from
        :return:
        """
        result = requests.get(url)
        if result.status_code != 200:
            raise Exception(f"Error fetching {url}: {result.status_code}")
        soup = BeautifulSoup(result.text)
        return soup.text



from __future__ import annotations
from datetime import datetime, date
from enum import Enum

from typing import List, Dict, Optional, Any, Union
from pydantic import BaseModel as BaseModel, ConfigDict,  Field, field_validator
import re
import sys
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


metamodel_version = "None"
version = "None"

class ConfiguredBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment=True,
        validate_default=True,
        extra = 'forbid',
        arbitrary_types_allowed=True,
        use_enum_values = True)
    pass


class NullDataOptions(str, Enum):
    
    
    UNSPECIFIED_METHOD_OF_ADMINISTRATION = "UNSPECIFIED_METHOD_OF_ADMINISTRATION"
    
    NOT_APPLICABLE = "NOT_APPLICABLE"
    
    NOT_MENTIONED = "NOT_MENTIONED"
    
    

class MaxoAnnotations(ConfiguredBaseModel):
    
    primary_disease: Optional[str] = Field(None, description="""The main disease the text is about, or its central disease topic. This is often the disease mentioned in an article's title or in its first few sentences.""")
    medical_actions: Optional[List[str]] = Field(default_factory=list, description="""Semicolon-separated list of medical actions.""")
    symptoms: Optional[List[str]] = Field(default_factory=list, description="""Semicolon-separated list of signs or symptoms.""")
    chemicals: Optional[List[str]] = Field(default_factory=list, description="""Semicolon-separated list of chemicals or drugs""")
    action_annotation_relationships: Optional[List[ActionAnnotationRelationship]] = Field(default_factory=list, description="""Semicolon-separated list of relationships between a disease, the mentioned signs and symptoms associated with that disease, the medical actions relating to each symptom, and the type of relationship between each action and symptom (usually TREATS or PREVENTS). The disease name must be included in the relationship, for example, \"treatment TREATS symptom IN disease\". If the medical action includes a specific chemical or drug, include the chemical or drug name in the relationship, for example, \"treatment (with chemical) TREATS symptom IN disease\".""")
    
    

class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    input_id: Optional[str] = Field(None)
    input_title: Optional[str] = Field(None)
    input_text: Optional[str] = Field(None)
    raw_completion_output: Optional[str] = Field(None)
    prompt: Optional[str] = Field(None)
    extracted_object: Optional[Any] = Field(None, description="""The complex objects extracted from the text""")
    named_entities: Optional[List[Any]] = Field(default_factory=list, description="""Named entities extracted from the text""")
    
    

class NamedEntity(ConfiguredBaseModel):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class MedicalAction(NamedEntity):
    """
    A clinically prescribed procedure, therapy, intervention, or recommendation. For example: blood transfusion, radiation therapy, cardiac catheterization, pulse oximetry, otoscopy
    """
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class Disease(NamedEntity):
    """
    A disposition to undergo pathological processes that exists in an organism because of one or more disorders in that organism. For example: Beck-Fahrner syndrome, hereditary retinoblastoma, progeria, diabetes mellitus, infectious otitis media
    """
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class Symptom(NamedEntity):
    """
    A condition or phenotype resulting from an abnormal health state. For example: Low serum calcitriol, hypoplasia of the thymus, chronic cough, aortic stiffness, low pulse pressure
    """
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class Chemical(NamedEntity):
    """
    A substance that has a defined molecular structure and is produced by or used in a chemical process. Includes drugs used as part of medical actions. For example: corticosteroid, folic acid, opioid analgesic
    """
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class CompoundExpression(ConfiguredBaseModel):
    
    None
    
    

class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""")
    
    

class ExtendedTriple(Triple):
    """
    Abstract parent for Relation Extraction tasks, with additional support for an extension term describing some aspect of the subject and object.
    """
    subject_extension: Optional[str] = Field(None, description="""An optional term describing some specific aspect of the subject, e.g. \"analgesic agent therapy\" has the aspect \"analgesic\"""")
    object_extension: Optional[str] = Field(None, description="""An optional term describing some specific aspect of the object, e.g. \"analgesic agent therapy\" has the aspect \"analgesic\"""")
    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""")
    
    

class ActionAnnotationRelationship(ExtendedTriple):
    """
    An association representing a relationships between a disease, the mentioned signs and symptoms associated with that disease, the medical actions relating to each symptom, and the type of relationship between each action and symptom (usually TREATS or PREVENTS).
    """
    subject_extension: Optional[str] = Field(None, description="""A chemical or drug mentioned in the relationship between the medical action and the symptom, for example, \"analgesic agent therapy\" has the aspect \"analgesic\"""")
    object_extension: Optional[str] = Field(None, description="""An optional term describing some specific aspect of the object, e.g. \"analgesic agent therapy\" has the aspect \"analgesic\"""")
    subject: Optional[str] = Field(None, description="""The medical action. For example: blood transfusion, radiation therapy, cardiac catheterization, pulse oximetry, otoscopy""")
    predicate: Optional[str] = Field(None, description="""The relationship type between the medical action and the symptom, usually TREATS or PREVENTS.""")
    object: Optional[str] = Field(None, description="""A sign or symptom associated with the disease and targeted by the medical action. For example, Low serum calcitriol, hypoplasia of the thymus, chronic cough, aortic stiffness, low pulse pressure""")
    qualifier: Optional[str] = Field(None, description="""The primary disease the relationship is about, or specifically the disease the symptom is related to. For example, Beck-Fahrner syndrome, hereditary retinoblastoma, progeria, diabetes mellitus, infectious otitis media""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""")
    
    

class TextWithTriples(ConfiguredBaseModel):
    """
    A text containing one or more relations of the Triple type.
    """
    publication: Optional[Publication] = Field(None)
    triples: Optional[List[Triple]] = Field(default_factory=list)
    
    

class TextWithEntity(ConfiguredBaseModel):
    """
    A text containing one or more instances of a single type of entity.
    """
    publication: Optional[Publication] = Field(None)
    entities: Optional[List[str]] = Field(default_factory=list)
    
    

class RelationshipType(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class Publication(ConfiguredBaseModel):
    
    id: Optional[str] = Field(None, description="""The publication identifier""")
    title: Optional[str] = Field(None, description="""The title of the publication""")
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""")
    combined_text: Optional[str] = Field(None)
    full_text: Optional[str] = Field(None, description="""The full text of the publication""")
    
    

class AnnotatorResult(ConfiguredBaseModel):
    
    subject_text: Optional[str] = Field(None)
    object_id: Optional[str] = Field(None)
    object_text: Optional[str] = Field(None)
    
    


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
MaxoAnnotations.model_rebuild()
ExtractionResult.model_rebuild()
NamedEntity.model_rebuild()
MedicalAction.model_rebuild()
Disease.model_rebuild()
Symptom.model_rebuild()
Chemical.model_rebuild()
CompoundExpression.model_rebuild()
Triple.model_rebuild()
ExtendedTriple.model_rebuild()
ActionAnnotationRelationship.model_rebuild()
TextWithTriples.model_rebuild()
TextWithEntity.model_rebuild()
RelationshipType.model_rebuild()
Publication.model_rebuild()
AnnotatorResult.model_rebuild()




from __future__ import annotations 

import re
import sys
from datetime import (
    date,
    datetime,
    time
)
from decimal import Decimal 
from enum import Enum 
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Literal,
    Optional,
    Union
)

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
    field_validator
)


metamodel_version = "None"
version = "None"


class ConfiguredBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment = True,
        validate_default = True,
        extra = "forbid",
        arbitrary_types_allowed = True,
        use_enum_values = True,
        strict = False,
    )
    pass




class LinkMLMeta(RootModel):
    root: Dict[str, Any] = {}
    model_config = ConfigDict(frozen=True)

    def __getattr__(self, key:str):
        return getattr(self.root, key)

    def __getitem__(self, key:str):
        return self.root[key]

    def __setitem__(self, key:str, value):
        self.root[key] = value

    def __contains__(self, key:str) -> bool:
        return key in self.root


linkml_meta = LinkMLMeta({'default_prefix': 'vbo_char',
     'default_range': 'string',
     'description': 'An extraction template for animal names present in VBO, along '
                    'with the characteristics of each breed',
     'id': 'http://w3id.org/ontogpt/vbo_char',
     'imports': ['linkml:types', 'core'],
     'license': 'https://creativecommons.org/publicdomain/zero/1.0/',
     'name': 'vbo_char',
     'prefixes': {'linkml': {'prefix_prefix': 'linkml',
                             'prefix_reference': 'https://w3id.org/linkml/'},
                  'rdf': {'prefix_prefix': 'rdf',
                          'prefix_reference': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'},
                  'vbo': {'prefix_prefix': 'vbo',
                          'prefix_reference': 'http://purl.obolibrary.org/obo/vbo'},
                  'vbo_char': {'prefix_prefix': 'vbo_char',
                               'prefix_reference': 'http://w3id.org/ontogpt/vbo_char'}},
     'source_file': '/home/harry/ontogpt/src/ontogpt/templates/vbo_char.yaml',
     'title': 'Extraction Template for Animal Breeds and their Characteristics'} )

class NullDataOptions(str, Enum):
    UNSPECIFIED_METHOD_OF_ADMINISTRATION = "UNSPECIFIED_METHOD_OF_ADMINISTRATION"
    NOT_APPLICABLE = "NOT_APPLICABLE"
    NOT_MENTIONED = "NOT_MENTIONED"



class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    input_id: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'input_id', 'domain_of': ['ExtractionResult']} })
    input_title: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'input_title', 'domain_of': ['ExtractionResult']} })
    input_text: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'input_text', 'domain_of': ['ExtractionResult']} })
    raw_completion_output: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'raw_completion_output', 'domain_of': ['ExtractionResult']} })
    prompt: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'prompt', 'domain_of': ['ExtractionResult']} })
    extracted_object: Optional[Any] = Field(None, description="""The complex objects extracted from the text""", json_schema_extra = { "linkml_meta": {'alias': 'extracted_object', 'domain_of': ['ExtractionResult']} })
    named_entities: Optional[List[Any]] = Field(None, description="""Named entities extracted from the text""", json_schema_extra = { "linkml_meta": {'alias': 'named_entities', 'domain_of': ['ExtractionResult']} })


class NamedEntity(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'abstract': True, 'from_schema': 'http://w3id.org/ontogpt/core'})

    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'annotations': {'owl': {'tag': 'owl',
                                 'value': 'AnnotationProperty, AnnotationAssertion'}},
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })
    original_spans: Optional[List[str]] = Field(None, description="""The coordinates of the original text span from which the named entity was extracted, inclusive. For example, \"10:25\" means the span starting from the 10th character and ending with the 25th character. The first character in the text has index 0. Newlines are treated as single characters. Multivalued as there may be multiple spans for a single text.""", json_schema_extra = { "linkml_meta": {'alias': 'original_spans',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['This is determined during grounding and normalization',
                      'But is based on the full input text'],
         'domain_of': ['NamedEntity']} })

    @field_validator('original_spans')
    def pattern_original_spans(cls, v):
        pattern=re.compile(r"^\d+:\d+$")
        if isinstance(v,list):
            for element in v:
                if isinstance(v, str) and not pattern.match(element):
                    raise ValueError(f"Invalid original_spans format: {element}")
        elif isinstance(v,str):
            if not pattern.match(v):
                raise ValueError(f"Invalid original_spans format: {v}")
        return v


class CompoundExpression(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'abstract': True, 'from_schema': 'http://w3id.org/ontogpt/core'})

    pass


class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'abstract': True, 'from_schema': 'http://w3id.org/ontogpt/core'})

    subject: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'subject', 'domain_of': ['Triple']} })
    predicate: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'predicate', 'domain_of': ['Triple']} })
    object: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'object', 'domain_of': ['Triple']} })
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""", json_schema_extra = { "linkml_meta": {'alias': 'qualifier', 'domain_of': ['Triple']} })
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""", json_schema_extra = { "linkml_meta": {'alias': 'subject_qualifier', 'domain_of': ['Triple']} })
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""", json_schema_extra = { "linkml_meta": {'alias': 'object_qualifier', 'domain_of': ['Triple']} })


class TextWithTriples(ConfiguredBaseModel):
    """
    A text containing one or more relations of the Triple type.
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    publication: Optional[Publication] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'publication',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'domain_of': ['TextWithTriples', 'TextWithEntity']} })
    triples: Optional[List[Triple]] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'triples', 'domain_of': ['TextWithTriples']} })


class TextWithEntity(ConfiguredBaseModel):
    """
    A text containing one or more instances of a single type of entity.
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    publication: Optional[Publication] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'publication',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'domain_of': ['TextWithTriples', 'TextWithEntity']} })
    entities: Optional[List[str]] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'entities', 'domain_of': ['TextWithEntity']} })


class RelationshipType(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core',
         'id_prefixes': ['RO', 'biolink']})

    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'annotations': {'owl': {'tag': 'owl',
                                 'value': 'AnnotationProperty, AnnotationAssertion'}},
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })
    original_spans: Optional[List[str]] = Field(None, description="""The coordinates of the original text span from which the named entity was extracted, inclusive. For example, \"10:25\" means the span starting from the 10th character and ending with the 25th character. The first character in the text has index 0. Newlines are treated as single characters. Multivalued as there may be multiple spans for a single text.""", json_schema_extra = { "linkml_meta": {'alias': 'original_spans',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['This is determined during grounding and normalization',
                      'But is based on the full input text'],
         'domain_of': ['NamedEntity']} })

    @field_validator('original_spans')
    def pattern_original_spans(cls, v):
        pattern=re.compile(r"^\d+:\d+$")
        if isinstance(v,list):
            for element in v:
                if isinstance(v, str) and not pattern.match(element):
                    raise ValueError(f"Invalid original_spans format: {element}")
        elif isinstance(v,str):
            if not pattern.match(v):
                raise ValueError(f"Invalid original_spans format: {v}")
        return v


class Publication(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    id: Optional[str] = Field(None, description="""The publication identifier""", json_schema_extra = { "linkml_meta": {'alias': 'id', 'domain_of': ['NamedEntity', 'Publication']} })
    title: Optional[str] = Field(None, description="""The title of the publication""", json_schema_extra = { "linkml_meta": {'alias': 'title', 'domain_of': ['Publication']} })
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""", json_schema_extra = { "linkml_meta": {'alias': 'abstract', 'domain_of': ['Publication']} })
    combined_text: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'combined_text', 'domain_of': ['Publication']} })
    full_text: Optional[str] = Field(None, description="""The full text of the publication""", json_schema_extra = { "linkml_meta": {'alias': 'full_text', 'domain_of': ['Publication']} })


class AnnotatorResult(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/core'})

    subject_text: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'subject_text', 'domain_of': ['AnnotatorResult']} })
    object_id: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'object_id', 'domain_of': ['AnnotatorResult']} })
    object_text: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'object_text', 'domain_of': ['AnnotatorResult']} })


class NameSet(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/vbo_char', 'tree_root': True})

    animal_names: Optional[List[str]] = Field(None, description="""A semicolon-separated list of names of animals used in the input text. These are general names, e.g., if any breed of cat is mentioned, this list should include \"Cat breed\", or for any pig, include \"Pig breed\".""", json_schema_extra = { "linkml_meta": {'alias': 'animal_names', 'domain_of': ['NameSet']} })
    names: Optional[List[str]] = Field(None, description="""A semicolon-separated list of names of animal breeds used in the input text. These should be as specific as possible about the breed of the animal. Examples include: Gimbsheimer Enten, Debao pony, Baixi""", json_schema_extra = { "linkml_meta": {'alias': 'names', 'domain_of': ['NameSet']} })
    characteristics: Optional[List[BreedToCharacteristic]] = Field(None, description="""A semicolon-separated list of names of animal breeds used in the input text, along with a single characteristic mentioned for that breed. These should be as specific as possible about the breed of the animal. The characteristic may be color, dimensions, physical properties, abilities, or other features. Each statement should contain just one pair of breed name and characteristic. It should be formatted as \"Breed IS Characteristic\", or \"Breed HAS Characteristic\", e.g., Gimbsheimer Enten IS blue, Debao pony IS short, Baixi IS circular Each additional characteristic for a breed should get its own statement, e.g., \"breed name is tall and wide\" should become \"Breed IS tall\" and \"Breed is wide\".""", json_schema_extra = { "linkml_meta": {'alias': 'characteristics', 'domain_of': ['NameSet']} })
    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'annotations': {'owl': {'tag': 'owl',
                                 'value': 'AnnotationProperty, AnnotationAssertion'}},
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })
    original_spans: Optional[List[str]] = Field(None, description="""The coordinates of the original text span from which the named entity was extracted, inclusive. For example, \"10:25\" means the span starting from the 10th character and ending with the 25th character. The first character in the text has index 0. Newlines are treated as single characters. Multivalued as there may be multiple spans for a single text.""", json_schema_extra = { "linkml_meta": {'alias': 'original_spans',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['This is determined during grounding and normalization',
                      'But is based on the full input text'],
         'domain_of': ['NamedEntity']} })

    @field_validator('original_spans')
    def pattern_original_spans(cls, v):
        pattern=re.compile(r"^\d+:\d+$")
        if isinstance(v,list):
            for element in v:
                if isinstance(v, str) and not pattern.match(element):
                    raise ValueError(f"Invalid original_spans format: {element}")
        elif isinstance(v,str):
            if not pattern.match(v):
                raise ValueError(f"Invalid original_spans format: {v}")
        return v


class AnimalName(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'annotations': {'annotators': {'tag': 'annotators', 'value': 'sqlite:obo:vbo'},
                         'prompt': {'tag': 'prompt',
                                    'value': 'The name of a grouping category of '
                                             'vertebrate animal breeds.'}},
         'from_schema': 'http://w3id.org/ontogpt/vbo_char'})

    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'annotations': {'owl': {'tag': 'owl',
                                 'value': 'AnnotationProperty, AnnotationAssertion'}},
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })
    original_spans: Optional[List[str]] = Field(None, description="""The coordinates of the original text span from which the named entity was extracted, inclusive. For example, \"10:25\" means the span starting from the 10th character and ending with the 25th character. The first character in the text has index 0. Newlines are treated as single characters. Multivalued as there may be multiple spans for a single text.""", json_schema_extra = { "linkml_meta": {'alias': 'original_spans',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['This is determined during grounding and normalization',
                      'But is based on the full input text'],
         'domain_of': ['NamedEntity']} })

    @field_validator('original_spans')
    def pattern_original_spans(cls, v):
        pattern=re.compile(r"^\d+:\d+$")
        if isinstance(v,list):
            for element in v:
                if isinstance(v, str) and not pattern.match(element):
                    raise ValueError(f"Invalid original_spans format: {element}")
        elif isinstance(v,str):
            if not pattern.match(v):
                raise ValueError(f"Invalid original_spans format: {v}")
        return v


class BreedName(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'annotations': {'annotators': {'tag': 'annotators', 'value': 'sqlite:obo:vbo'},
                         'prompt': {'tag': 'prompt',
                                    'value': 'The name of a vertebrate animal breed.'}},
         'from_schema': 'http://w3id.org/ontogpt/vbo_char'})

    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'annotations': {'owl': {'tag': 'owl',
                                 'value': 'AnnotationProperty, AnnotationAssertion'}},
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })
    original_spans: Optional[List[str]] = Field(None, description="""The coordinates of the original text span from which the named entity was extracted, inclusive. For example, \"10:25\" means the span starting from the 10th character and ending with the 25th character. The first character in the text has index 0. Newlines are treated as single characters. Multivalued as there may be multiple spans for a single text.""", json_schema_extra = { "linkml_meta": {'alias': 'original_spans',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['This is determined during grounding and normalization',
                      'But is based on the full input text'],
         'domain_of': ['NamedEntity']} })

    @field_validator('original_spans')
    def pattern_original_spans(cls, v):
        pattern=re.compile(r"^\d+:\d+$")
        if isinstance(v,list):
            for element in v:
                if isinstance(v, str) and not pattern.match(element):
                    raise ValueError(f"Invalid original_spans format: {element}")
        elif isinstance(v,str):
            if not pattern.match(v):
                raise ValueError(f"Invalid original_spans format: {v}")
        return v


class Characteristic(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'annotations': {'annotators': {'tag': 'annotators',
                                        'value': 'sqlite:obo:uberon, sqlite:obo:pato'},
                         'prompt': {'tag': 'prompt',
                                    'value': 'The name of a characteristic of an '
                                             'animal.'}},
         'from_schema': 'http://w3id.org/ontogpt/vbo_char'})

    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'annotations': {'owl': {'tag': 'owl',
                                 'value': 'AnnotationProperty, AnnotationAssertion'}},
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })
    original_spans: Optional[List[str]] = Field(None, description="""The coordinates of the original text span from which the named entity was extracted, inclusive. For example, \"10:25\" means the span starting from the 10th character and ending with the 25th character. The first character in the text has index 0. Newlines are treated as single characters. Multivalued as there may be multiple spans for a single text.""", json_schema_extra = { "linkml_meta": {'alias': 'original_spans',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['This is determined during grounding and normalization',
                      'But is based on the full input text'],
         'domain_of': ['NamedEntity']} })

    @field_validator('original_spans')
    def pattern_original_spans(cls, v):
        pattern=re.compile(r"^\d+:\d+$")
        if isinstance(v,list):
            for element in v:
                if isinstance(v, str) and not pattern.match(element):
                    raise ValueError(f"Invalid original_spans format: {element}")
        elif isinstance(v,str):
            if not pattern.match(v):
                raise ValueError(f"Invalid original_spans format: {v}")
        return v


class Descriptor(NamedEntity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'annotations': {'annotators': {'tag': 'annotators',
                                        'value': 'sqlite:obo:pato'},
                         'prompt': {'tag': 'prompt',
                                    'value': 'A descriptor for a characteristic.'}},
         'from_schema': 'http://w3id.org/ontogpt/vbo_char'})

    id: str = Field(..., description="""A unique identifier for the named entity""", json_schema_extra = { "linkml_meta": {'alias': 'id',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['this is populated during the grounding and normalization step'],
         'domain_of': ['NamedEntity', 'Publication']} })
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""", json_schema_extra = { "linkml_meta": {'alias': 'label',
         'aliases': ['name'],
         'annotations': {'owl': {'tag': 'owl',
                                 'value': 'AnnotationProperty, AnnotationAssertion'}},
         'domain_of': ['NamedEntity'],
         'slot_uri': 'rdfs:label'} })
    original_spans: Optional[List[str]] = Field(None, description="""The coordinates of the original text span from which the named entity was extracted, inclusive. For example, \"10:25\" means the span starting from the 10th character and ending with the 25th character. The first character in the text has index 0. Newlines are treated as single characters. Multivalued as there may be multiple spans for a single text.""", json_schema_extra = { "linkml_meta": {'alias': 'original_spans',
         'annotations': {'prompt.skip': {'tag': 'prompt.skip', 'value': 'true'}},
         'comments': ['This is determined during grounding and normalization',
                      'But is based on the full input text'],
         'domain_of': ['NamedEntity']} })

    @field_validator('original_spans')
    def pattern_original_spans(cls, v):
        pattern=re.compile(r"^\d+:\d+$")
        if isinstance(v,list):
            for element in v:
                if isinstance(v, str) and not pattern.match(element):
                    raise ValueError(f"Invalid original_spans format: {element}")
        elif isinstance(v,str):
            if not pattern.match(v):
                raise ValueError(f"Invalid original_spans format: {v}")
        return v


class BreedToCharacteristic(Triple):
    """
    A triple in which the subject is an animal breed, the object is a characteristic, and the predicate is usually \"IS\" or \"HAS\".
    """
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({'from_schema': 'http://w3id.org/ontogpt/vbo_char',
         'slot_usage': {'object': {'description': 'The specific characteristic. This '
                                                  'is generally a noun of the '
                                                  'characteristic, e.g., with "blue '
                                                  'fin" the object is "fin".',
                                   'name': 'object',
                                   'range': 'Characteristic'},
                        'object_qualifier': {'description': 'An optional qualifier or '
                                                            'modifier for the '
                                                            'characteristic. This is '
                                                            'generally a descriptor of '
                                                            'the characteristic, e.g., '
                                                            'with "blue fin" the '
                                                            'qualifier is "blue".',
                                             'name': 'object_qualifier',
                                             'range': 'Descriptor'},
                        'predicate': {'description': 'The relationship type, generally '
                                                     'IS or HAS to indicate a breed is '
                                                     'defined by having a specific '
                                                     'characteristic.',
                                      'name': 'predicate',
                                      'range': 'NamedEntity'},
                        'subject': {'name': 'subject', 'range': 'BreedName'},
                        'subject_qualifier': {'description': 'An optional qualifier or '
                                                             'modifier for the breed.',
                                              'name': 'subject_qualifier',
                                              'range': 'NamedEntity'}}})

    subject: Optional[str] = Field(None, json_schema_extra = { "linkml_meta": {'alias': 'subject', 'domain_of': ['Triple']} })
    predicate: Optional[str] = Field(None, description="""The relationship type, generally IS or HAS to indicate a breed is defined by having a specific characteristic.""", json_schema_extra = { "linkml_meta": {'alias': 'predicate', 'domain_of': ['Triple']} })
    object: Optional[str] = Field(None, description="""The specific characteristic. This is generally a noun of the characteristic, e.g., with \"blue fin\" the object is \"fin\".""", json_schema_extra = { "linkml_meta": {'alias': 'object', 'domain_of': ['Triple']} })
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""", json_schema_extra = { "linkml_meta": {'alias': 'qualifier', 'domain_of': ['Triple']} })
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the breed.""", json_schema_extra = { "linkml_meta": {'alias': 'subject_qualifier', 'domain_of': ['Triple']} })
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the characteristic. This is generally a descriptor of the characteristic, e.g., with \"blue fin\" the qualifier is \"blue\".""", json_schema_extra = { "linkml_meta": {'alias': 'object_qualifier', 'domain_of': ['Triple']} })


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
ExtractionResult.model_rebuild()
NamedEntity.model_rebuild()
CompoundExpression.model_rebuild()
Triple.model_rebuild()
TextWithTriples.model_rebuild()
TextWithEntity.model_rebuild()
RelationshipType.model_rebuild()
Publication.model_rebuild()
AnnotatorResult.model_rebuild()
NameSet.model_rebuild()
AnimalName.model_rebuild()
BreedName.model_rebuild()
Characteristic.model_rebuild()
Descriptor.model_rebuild()
BreedToCharacteristic.model_rebuild()
